import logging

import numpy as np

from pylops import LinearOperator

try:
    from numba import jit

    from ._Spread_numba import (
        _matvec_numba_onthefly,
        _matvec_numba_table,
        _rmatvec_numba_onthefly,
        _rmatvec_numba_table,
    )
except ModuleNotFoundError:
    jit = None
    jit_message = "Numba not available, reverting to numpy."
except Exception as e:
    jit = None
    jit_message = "Failed to import numba (error:%s), use numpy." % e

logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.WARNING)


class Spread(LinearOperator):
    r"""Spread operator.

    Spread values from the input model vector arranged as a 2-dimensional
    array of size :math:`[n_{x_0} \times n_{t_0}]` into the data vector of size
    :math:`[n_x \times n_t]`. Note that the value at each single pair
    :math:`(x_0, t_0)` in the input is spread over the entire :math:`x` axis
    in the output.

    Spreading is performed along parametric curves provided as look-up table
    of pre-computed indices (``table``) or computed on-the-fly using a
    function handle (``fh``).

    In adjont mode, values from the data vector are instead stacked
    along the same parametric curves.

    Parameters
    ----------
    dims : :obj:`tuple`
        Dimensions of model vector (vector will be reshaped internally into
        a two-dimensional array of size :math:`[n_{x_0} \times n_{t_0}]`,
        where the first dimension is the spreading direction)
    dimsd : :obj:`tuple`
        Dimensions of data vector (vector will be reshaped internal into
        a two-dimensional array of size :math:`[n_x \times n_t]`,
        where the first dimension is the stacking direction)
    table : :obj:`np.ndarray`, optional
        Look-up table of indices of size
        :math:`[n_{x_0} \times n_{t_0} \times n_x]` (if ``None`` use function
        handle ``fh``). When ``dtable`` is not provided, the  ``data`` will be created
        as follows

        .. code-block:: python

            data[ix, table[ix0, it0, ix]] += model[ix0, it0]

        .. note::
            When using ``table`` without ``dtable``, its elements must be
            between 0 and :math:`n_{t_0} - 1` (or ``numpy.nan``).

    dtable : :obj:`np.ndarray`, optional
        Look-up table of decimals remainders for linear interpolation of size
        :math:`[n_{x_0} \times n_{t_0} \times n_x]` (if ``None`` use function
        handle ``fh``). When provided, the ``data`` will be created as follows

        .. code-block:: python

            data[ix, table[ix0, it0, ix]]     += (1 - dtable[ix0, it0, ix]) * model[ix0, it0]
            data[ix, table[ix0, it0, ix] + 1] +=      dtable[ix0, it0, ix]  * model[ix0, it0]

        .. note::
            When using ``table`` and ``dtable``, the elements of ``table`` indices must be
            between 0 and :math:`n_{t_0} - 2`  (or ``numpy.nan``).

    fh : :obj:`callable`, optional
        If ``None`` will use look-up table ``table``. When provided, should be a
        function which takes indices ``ix0`` and ``it0`` and returns
        an array of size :math:`n_x` containing each respective time index.
        Alternatively, if linear interpolation is required, it should output in
        addition to the time indices, a weight for interpolation with linear
        interpolation, to be used as follows

        .. code-block:: python

            data[ix, index]     += (1 - dindices[ix]) * model[ix0, it0]
            data[ix, index + 1] +=      dindices[ix]  * model[ix0, it0]

        where ``index`` refers to a time index in the first array returned by ``fh``
        and ``dindices`` refers to the weight in the second array returned by ``fh``.

        .. note::
            When using ``fh`` with one output (time indices), the time indices must be
            between 0 and :math:`n_{t_0} - 1` (or ``numpy.nan``). When using ``fh`` with two outputs
            (time indices and weights), they must be within the between 0 and
            :math:`n_{t_0} - 2` (or ``numpy.nan``).

    interp : :obj:`bool`, optional
        Use only if engine ``engine='numba'``. Apply linear interpolation (``True``)
        or nearest interpolation (``False``) during stacking/spreading along
        parametric curve.
        When using ``engine="numpy"``, it will be inferred directly from ``fh`` or
        the presence of ``dtable``.
    engine : :obj:`str`, optional
        Engine used for fft computation (``numpy`` or ``numba``). Note that
        ``numba`` can only be used when providing a look-up table
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

    Raises
    ------
    KeyError
        If ``engine`` is neither ``numpy`` nor ``numba``
    NotImplementedError
        If both ``table`` and ``fh`` are not provided
    ValueError
        If ``table`` has shape different from
        :math:`[n_{x_0} \times n_{t_0} \times n_x]`

    Notes
    -----
    The Spread operator applies the following linear transform in forward mode
    to the model vector after reshaping it into a 2-dimensional array of size
    :math:`[n_x \times n_t]`:

    .. math::
        m(x_0, t_0) \rightarrow d(x, t=f(x_0, x, t_0)) \quad \forall x

    where :math:`f(x_0, x, t)` is a mapping function that returns a value :math:`t`
    given values :math:`x_0`, :math:`x`, and  :math:`t_0`. Note that for each
    :math:`(x_0, t_0)` pair, spreading is done over the entire :math:`x` axis
    in the data domain.

    In adjoint mode, the model is reconstructed by means of the following
    stacking operation:

    .. math::
        m(x_0, t_0) = \int{d(x, t=f(x_0, x, t_0))} \,\mathrm{d}x

    Note that ``table`` (or ``fh``)  must return integer numbers
    representing indices in the axis :math:`t`. However it also possible to
    perform linear interpolation as part of the spreading/stacking process by
    providing the decimal part of the mapping function (:math:`t - \lfloor
    t \rfloor`) either in ``dtable`` input parameter or as second value in
    the return of ``fh`` function.

    """

    def __init__(
        self,
        dims,
        dimsd,
        table=None,
        dtable=None,
        fh=None,
        interp=None,
        engine="numpy",
        dtype="float64",
    ):
        if engine not in ["numpy", "numba"]:
            raise KeyError("engine must be numpy or numba")
        if engine == "numba" and jit is not None:
            self.engine = "numba"
        else:
            if engine == "numba" and jit is None:
                logging.warning(jit_message)
            self.engine = "numpy"

        # axes
        self.dims, self.dimsd = dims, dimsd
        self.nx0, self.nt0 = self.dims[0], self.dims[1]
        self.nx, self.nt = self.dimsd[0], self.dimsd[1]
        self.table = table
        self.dtable = dtable
        self.fh = fh

        # find out if mapping is in table of function handle
        if table is None and fh is None:
            raise NotImplementedError("provide either table or fh.")
        elif table is not None:
            if fh is not None:
                raise ValueError("provide only one of table or fh.")
            if self.table.shape != (self.nx0, self.nt0, self.nx):
                raise ValueError("table must have shape [nx0 x nt0 x nx]")
            self.usetable = True
            if np.any(self.table > self.nt):
                raise ValueError("values in table must be smaller than nt")
        else:
            self.usetable = False

        # find out if linear interpolation has to be carried out
        self.interp = False
        if self.usetable:
            if dtable is not None:
                if self.dtable.shape != (self.nx0, self.nt0, self.nx):
                    raise ValueError("dtable must have shape [nx0 x nt x nx]")
                self.interp = True
        else:
            if self.engine == "numba":
                self.interp = interp
            else:
                if len(fh(0, 0)) == 2:
                    self.interp = True
        if interp is not None and self.interp != interp:
            logging.warning("interp has been overridden to %r.", self.interp)
        self.shape = (int(np.prod(self.dimsd)), int(np.prod(self.dims)))
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec_numpy(self, x):
        x = x.reshape(self.dims)
        y = np.zeros(self.dimsd, dtype=self.dtype)
        for it in range(self.dims[1]):
            for ix0 in range(self.dims[0]):
                if self.usetable:
                    indices = self.table[ix0, it]
                    if self.interp:
                        dindices = self.dtable[ix0, it]
                else:
                    if self.interp:
                        indices, dindices = self.fh(ix0, it)
                    else:
                        indices = self.fh(ix0, it)
                mask = np.argwhere(~np.isnan(indices))
                if mask.size > 0:
                    indices = (indices[mask]).astype(int)
                    if not self.interp:
                        y[mask, indices] += x[ix0, it]
                    else:
                        y[mask, indices] += (1 - dindices[mask]) * x[ix0, it]
                        y[mask, indices + 1] += dindices[mask] * x[ix0, it]
        return y.ravel()

    def _rmatvec_numpy(self, x):
        x = x.reshape(self.dimsd)
        y = np.zeros(self.dims, dtype=self.dtype)
        for it in range(self.dims[1]):
            for ix0 in range(self.dims[0]):
                if self.usetable:
                    indices = self.table[ix0, it]
                    if self.interp:
                        dindices = self.dtable[ix0, it]
                else:
                    if self.interp:
                        indices, dindices = self.fh(ix0, it)
                    else:
                        indices = self.fh(ix0, it)
                mask = np.argwhere(~np.isnan(indices))
                if mask.size > 0:
                    indices = (indices[mask]).astype(int)
                    if not self.interp:
                        y[ix0, it] = np.sum(x[mask, indices])
                    else:
                        y[ix0, it] = np.sum(
                            x[mask, indices] * (1 - dindices[mask])
                        ) + np.sum(x[mask, indices + 1] * dindices[mask])
        return y.ravel()

    def _matvec(self, x):
        if self.engine == "numba":
            y = np.zeros(self.dimsd, dtype=self.dtype)
            if self.usetable:
                y = _matvec_numba_table(
                    x,
                    y,
                    self.dims,
                    self.interp,
                    self.table,
                    self.table if self.dtable is None else self.dtable,
                )
            else:
                y = _matvec_numba_onthefly(x, y, self.dims, self.interp, self.fh)
        else:
            y = self._matvec_numpy(x)
        return y

    def _rmatvec(self, x):
        if self.engine == "numba":
            y = np.zeros(self.dims, dtype=self.dtype)
            if self.usetable:
                y = _rmatvec_numba_table(
                    x,
                    y,
                    self.dims,
                    self.dimsd,
                    self.interp,
                    self.table,
                    self.table if self.dtable is None else self.dtable,
                )
            else:
                y = _rmatvec_numba_onthefly(
                    x, y, self.dims, self.dimsd, self.interp, self.fh
                )
        else:
            y = self._rmatvec_numpy(x)
        return y



import logging

import numpy as np

from pylops.basicoperators import BlockDiag, Diagonal, HStack, Restriction
from pylops.signalprocessing.Sliding2D import _slidingsteps
from pylops.utils.tapers import taper2d

logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.WARNING)


def Patch2D(Op, dims, dimsd, nwin, nover, nop, tapertype="hanning", design=False):
    """2D Patch transform operator.

    Apply a transform operator ``Op`` repeatedly to patches of the model
    vector in forward mode and patches of the data vector in adjoint mode.
    More specifically, in forward mode the model vector is divided into
    patches, each patch is transformed, and patches are then recombined
    together. Both model and data are internally reshaped and
    interpreted as 2-dimensional arrays: each patch contains a portion
    of the array in both the first and second dimension.

    This operator can be used to perform local, overlapping transforms (e.g.,
    :obj:`pylops.signalprocessing.FFT2D`
    or :obj:`pylops.signalprocessing.Radon2D`) on 2-dimensional arrays.

    .. note:: The shape of the model has to be consistent with
       the number of windows for this operator not to return an error. As the
       number of windows depends directly on the choice of ``nwin`` and
       ``nover``, it is recommended to use ``design=True`` if unsure about the
       choice ``dims`` and use the number of windows printed on screen to
       define such input parameter.

    .. warning:: Depending on the choice of `nwin` and `nover` as well as the
       size of the data, patches may not cover the entire size of the data.
       The start and end indices of each window can be displayed using
       ``design=True`` while defining the best patching approach.

    Parameters
    ----------
    Op : :obj:`pylops.LinearOperator`
        Transform operator
    dims : :obj:`tuple`
        Shape of 2-dimensional model. Note that ``dims[0]`` and ``dims[1]``
        should be multiple of the model size of the transform in their
        respective dimensions
    dimsd : :obj:`tuple`
        Shape of 2-dimensional data
    nwin : :obj:`tuple`
        Number of samples of window
    nover : :obj:`tuple`
        Number of samples of overlapping part of window
    nop : :obj:`tuple`
        Size of model in the transformed domain
    tapertype : :obj:`str`, optional
        Type of taper (``hanning``, ``cosine``, ``cosinesquare`` or ``None``)
    design : :obj:`bool`, optional
        Print number of sliding window (``True``) or not (``False``)

    Returns
    -------
    Sop : :obj:`pylops.LinearOperator`
        Sliding operator

    Raises
    ------
    ValueError
        Identified number of windows is not consistent with provided model
        shape (``dims``).

    See Also
    --------
    Sliding2d: 2D Sliding transform operator.

    """
    # model windows
    mwin0_ins, mwin0_ends = _slidingsteps(dims[0], nop[0], 0)
    mwin1_ins, mwin1_ends = _slidingsteps(dims[1], nop[1], 0)

    # data windows
    dwin0_ins, dwin0_ends = _slidingsteps(dimsd[0], nwin[0], nover[0])
    dwin1_ins, dwin1_ends = _slidingsteps(dimsd[1], nwin[1], nover[1])
    nwins0 = len(dwin0_ins)
    nwins1 = len(dwin1_ins)
    nwins = nwins0 * nwins1

    # create tapers
    if tapertype is not None:
        tap = taper2d(nwin[1], nwin[0], nover, tapertype=tapertype).astype(Op.dtype)
        taps = {itap: tap for itap in range(nwins)}
        # topmost tapers
        taptop = tap.copy()
        taptop[: nover[0]] = tap[nwin[0] // 2]
        for itap in range(0, nwins1):
            taps[itap] = taptop
        # bottommost tapers
        tapbottom = tap.copy()
        tapbottom[-nover[0] :] = tap[nwin[0] // 2]
        for itap in range(nwins - nwins1, nwins):
            taps[itap] = tapbottom
        # leftmost tapers
        tapleft = tap.copy()
        tapleft[:, : nover[1]] = tap[:, nwin[1] // 2][:, np.newaxis]
        for itap in range(0, nwins, nwins1):
            taps[itap] = tapleft
        # rightmost tapers
        tapright = tap.copy()
        tapright[:, -nover[1] :] = tap[:, nwin[1] // 2][:, np.newaxis]
        for itap in range(nwins1 - 1, nwins, nwins1):
            taps[itap] = tapright
        # lefttopcorner taper
        taplefttop = tap.copy()
        taplefttop[:, : nover[1]] = tap[:, nwin[1] // 2][:, np.newaxis]
        taplefttop[: nover[0]] = taplefttop[nwin[0] // 2]
        taps[0] = taplefttop
        # righttopcorner taper
        taprighttop = tap.copy()
        taprighttop[:, -nover[1] :] = tap[:, nwin[1] // 2][:, np.newaxis]
        taprighttop[: nover[0]] = taprighttop[nwin[0] // 2]
        taps[nwins1 - 1] = taprighttop
        # leftbottomcorner taper
        tapleftbottom = tap.copy()
        tapleftbottom[:, : nover[1]] = tap[:, nwin[1] // 2][:, np.newaxis]
        tapleftbottom[-nover[0] :] = tapleftbottom[nwin[0] // 2]
        taps[nwins - nwins1] = tapleftbottom
        # rightbottomcorner taper
        taprightbottom = tap.copy()
        taprightbottom[:, -nover[1] :] = tap[:, nwin[1] // 2][:, np.newaxis]
        taprightbottom[-nover[0] :] = taprightbottom[nwin[0] // 2]
        taps[nwins - 1] = taprightbottom

    # check that identified number of windows agrees with mode size
    if design:
        logging.warning("%d-%d windows required...", nwins0, nwins1)
        logging.warning(
            "model wins - start:%s, end:%s / start:%s, end:%s",
            str(mwin0_ins),
            str(mwin0_ends),
            str(mwin1_ins),
            str(mwin1_ends),
        )
        logging.warning(
            "data wins - start:%s, end:%s / start:%s, end:%s",
            str(dwin0_ins),
            str(dwin0_ends),
            str(dwin1_ins),
            str(dwin1_ends),
        )
    if nwins0 * nop[0] != dims[0] or nwins1 * nop[1] != dims[1]:
        raise ValueError(
            "Model shape (dims=%s) is not consistent with chosen "
            "number of windows. Choose dims[0]=%d and "
            "dims[1]=%d for the operator to work with "
            "estimated number of windows, or create "
            "the operator with design=True to find out the"
            "optimal number of windows for the current "
            "model size..." % (str(dims), nwins0 * nop[0], nwins1 * nop[1])
        )
    # transform to apply
    if tapertype is None:
        OOp = BlockDiag([Op for _ in range(nwins)])
    else:
        OOp = BlockDiag(
            [Diagonal(taps[itap].ravel(), dtype=Op.dtype) * Op for itap in range(nwins)]
        )

    hstack = HStack(
        [
            Restriction(
                dimsd[1] * nwin[0],
                range(win_in, win_end),
                dims=(nwin[0], dimsd[1]),
                dir=1,
                dtype=Op.dtype,
            ).H
            for win_in, win_end in zip(dwin1_ins, dwin1_ends)
        ]
    )

    combining1 = BlockDiag([hstack] * nwins0)
    combining0 = HStack(
        [
            Restriction(
                np.prod(dimsd),
                range(win_in, win_end),
                dims=dimsd,
                dir=0,
                dtype=Op.dtype,
            ).H
            for win_in, win_end in zip(dwin0_ins, dwin0_ends)
        ]
    )
    Pop = combining0 * combining1 * OOp
    return Pop



import numpy as np

from pylops import LinearOperator


class CausalIntegration(LinearOperator):
    r"""Causal integration.

    Apply causal integration to a multi-dimensional array along ``dir`` axis.

    Parameters
    ----------
    N : :obj:`int`
        Number of samples in model.
    dims : :obj:`list`, optional
        Number of samples for each dimension
        (``None`` if only one dimension is available)
    dir : :obj:`int`, optional
        Direction along which smoothing is applied.
    sampling : :obj:`float`, optional
        Sampling step ``dx``.
    halfcurrent : :obj:`bool`, optional
        Add half of current value (``True``) or the entire value (``False``).
        This will be *deprecated* in v2.0.0, use instead `kind=half` to obtain
        the same behaviour.
    dtype : :obj:`str`, optional
        Type of elements in input array.
    kind : :obj:`str`, optional
        Integration kind (``full``, ``half``, or ``trapezoidal``).
    removefirst : :obj:`bool`, optional
        Remove first sample (``True``) or not (``False``).

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``)
        or not (``False``)

    Notes
    -----
    The CausalIntegration operator applies a causal integration to any chosen
    direction of a multi-dimensional array.

    For simplicity, given a one dimensional array, the causal integration is:

    .. math::
        y(t) = \int\limits_{-\infty}^t x(\tau) \,\mathrm{d}\tau

    which can be discretised as :

    .. math::
        y[i] = \sum_{j=0}^i x[j] \,\Delta t

    or

    .. math::
        y[i] = \left(\sum_{j=0}^{i-1} x[j] + 0.5x[i]\right) \,\Delta t

    or

    .. math::
        y[i] = \left(\sum_{j=1}^{i-1} x[j] + 0.5x[0] + 0.5x[i]\right) \,\Delta t

    where :math:`\Delta t` is the ``sampling`` interval, and assuming the signal is zero
    before sample :math:`j=0`. In our implementation, the
    choice to add :math:`x[i]` or :math:`0.5x[i]` is made by selecting ``kind=full``
    or ``kind=half``, respectively. The choice to add :math:`0.5x[i]` and
    :math:`0.5x[0]` instead of made by selecting the ``kind=trapezoidal``.

    Note that the causal integral of a signal will depend, up to a constant,
    on causal start of the signal. For example if :math:`x(\tau) = t^2` the
    resulting indefinite integration is:

    .. math::
        y(t) = \int \tau^2 \,\mathrm{d}\tau = \frac{t^3}{3} + C

    However, if we apply a first derivative to :math:`y` always obtain:

    .. math::
        x(t) = \frac{\mathrm{d}y}{\mathrm{d}t} = t^2

    no matter the choice of :math:`C`.

    """

    def __init__(
        self,
        N,
        dims=None,
        dir=-1,
        sampling=1,
        halfcurrent=True,
        dtype="float64",
        kind="full",
        removefirst=False,
    ):
        self.N = N
        self.dir = dir
        self.sampling = sampling
        self.kind = kind
        if kind == "full" and halfcurrent:  # ensure backcompatibility
            self.kind = "half"
        self.removefirst = removefirst
        # define samples to remove from output
        rf = 0
        if removefirst:
            rf = 1 if dims is None else self.N // dims[self.dir]
        if dims is None:
            self.dims = [self.N, 1]
            self.dimsd = [self.N - rf, 1]
            self.reshape = False
        else:
            if np.prod(dims) != self.N:
                raise ValueError("product of dims must equal N!")
            else:
                self.dims = dims
                self.dimsd = list(dims)
                if self.removefirst:
                    self.dimsd[self.dir] -= 1
                self.reshape = True
        self.shape = (self.N - rf, self.N)
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        if self.reshape:
            x = np.reshape(x, self.dims)
        if self.dir != -1:
            x = np.swapaxes(x, self.dir, -1)
        y = self.sampling * np.cumsum(x, axis=-1)
        if self.kind in ("half", "trapezoidal"):
            y -= self.sampling * x / 2.0
        if self.kind == "trapezoidal":
            y[..., 1:] -= self.sampling * x[..., 0:1] / 2.0
        if self.removefirst:
            y = y[..., 1:]
        if self.dir != -1:
            y = np.swapaxes(y, -1, self.dir)
        return y.ravel()

    def _rmatvec(self, x):
        if self.reshape:
            x = np.reshape(x, self.dimsd)
        if self.removefirst:
            x = np.insert(x, 0, 0, axis=self.dir)
        if self.dir != -1:
            x = np.swapaxes(x, self.dir, -1)
        xflip = np.flip(x, axis=-1)
        if self.kind == "half":
            y = self.sampling * (np.cumsum(xflip, axis=-1) - xflip / 2.0)
        elif self.kind == "trapezoidal":
            y = self.sampling * (np.cumsum(xflip, axis=-1) - xflip / 2.0)
            y[..., -1] = self.sampling * np.sum(xflip, axis=-1) / 2.0
        else:
            y = self.sampling * np.cumsum(xflip, axis=-1)
        y = np.flip(y, axis=-1)
        if self.dir != -1:
            y = np.swapaxes(y, -1, self.dir)
        return y.ravel()



# isort: skip_file
from .lsm import LSM, Demigration
from .marchenko import Marchenko
from .mdd import MDC, MDD
from .oneway import Deghosting, PhaseShift
from .seismicinterpolation import SeismicInterpolation
from .wavedecomposition import (
    PressureToVelocity,
    UpDownComposition2D,
    UpDownComposition3D,
    WavefieldDecomposition,
)



import multiprocessing as mp

import numpy as np
from scipy.sparse.linalg.interface import LinearOperator as spLinearOperator
from scipy.sparse.linalg.interface import _get_dtype

from pylops import LinearOperator
from pylops.basicoperators import MatrixMult
from pylops.utils.backend import get_array_module


def _matvec_rmatvec_map(op, x):
    """matvec/rmatvec for multiprocessing"""
    return op(x).squeeze()


class BlockDiag(LinearOperator):
    r"""Block-diagonal operator.

    Create a block-diagonal operator from N linear operators.

    Parameters
    ----------
    ops : :obj:`list`
        Linear operators to be stacked. Alternatively,
        :obj:`numpy.ndarray` or :obj:`scipy.sparse` matrices can be passed
        in place of one or more operators.
    nproc : :obj:`int`, optional
        Number of processes used to evaluate the N operators in parallel using
        ``multiprocessing``. If ``nproc=1``, work in serial mode.
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

    Notes
    -----
    A block-diagonal operator composed of N linear operators is created such
    as its application in forward mode leads to

    .. math::
        \begin{bmatrix}
            \mathbf{L}_1  & \mathbf{0}   &  \ldots &  \mathbf{0}  \\
            \mathbf{0}    & \mathbf{L}_2 &  \ldots &  \mathbf{0}  \\
            \vdots        & \vdots       &  \ddots &  \vdots         \\
            \mathbf{0}    & \mathbf{0}   &  \ldots &  \mathbf{L}_N
        \end{bmatrix}
        \begin{bmatrix}
            \mathbf{x}_{1}  \\
            \mathbf{x}_{2}  \\
            \vdots     \\
            \mathbf{x}_{N}
        \end{bmatrix} =
        \begin{bmatrix}
            \mathbf{L}_1 \mathbf{x}_{1}  \\
            \mathbf{L}_2 \mathbf{x}_{2}  \\
            \vdots     \\
            \mathbf{L}_N \mathbf{x}_{N}
        \end{bmatrix}

    while its application in adjoint mode leads to

    .. math::
        \begin{bmatrix}
            \mathbf{L}_1^H  & \mathbf{0}     & \ldots & \mathbf{0}  \\
            \mathbf{0}      & \mathbf{L}_2^H & \ldots & \mathbf{0}  \\
            \vdots          & \vdots         & \ddots & \vdots      \\
            \mathbf{0}      & \mathbf{0}     & \ldots & \mathbf{L}_N^H
        \end{bmatrix}
        \begin{bmatrix}
            \mathbf{y}_{1}  \\
            \mathbf{y}_{2}  \\
            \vdots     \\
            \mathbf{y}_{N}
        \end{bmatrix} =
        \begin{bmatrix}
            \mathbf{L}_1^H \mathbf{y}_{1}  \\
            \mathbf{L}_2^H \mathbf{y}_{2}  \\
            \vdots     \\
            \mathbf{L}_N^H \mathbf{y}_{N}
        \end{bmatrix}

    """

    def __init__(self, ops, nproc=1, dtype=None):
        self.ops = ops
        mops = np.zeros(len(ops), dtype=int)
        nops = np.zeros(len(ops), dtype=int)
        for iop, oper in enumerate(ops):
            if not isinstance(oper, (LinearOperator, spLinearOperator)):
                self.ops[iop] = MatrixMult(oper, dtype=oper.dtype)
            nops[iop] = self.ops[iop].shape[0]
            mops[iop] = self.ops[iop].shape[1]
        self.nops = int(nops.sum())
        self.mops = int(mops.sum())
        self.nnops = np.insert(np.cumsum(nops), 0, 0)
        self.mmops = np.insert(np.cumsum(mops), 0, 0)
        # create pool for multiprocessing
        self._nproc = nproc
        self.pool = None
        if self.nproc > 1:
            self.pool = mp.Pool(processes=nproc)
        self.shape = (self.nops, self.mops)
        if dtype is None:
            self.dtype = _get_dtype(ops)
        else:
            self.dtype = np.dtype(dtype)
        self.explicit = False
        self.clinear = all([getattr(oper, "clinear", True) for oper in self.ops])

    @property
    def nproc(self):
        return self._nproc

    @nproc.setter
    def nproc(self, nprocnew):
        if self._nproc > 1:
            self.pool.close()
        if nprocnew > 1:
            self.pool = mp.Pool(processes=nprocnew)
        self._nproc = nprocnew

    def _matvec_serial(self, x):
        ncp = get_array_module(x)
        y = ncp.zeros(self.nops, dtype=self.dtype)
        for iop, oper in enumerate(self.ops):
            y[self.nnops[iop] : self.nnops[iop + 1]] = oper.matvec(
                x[self.mmops[iop] : self.mmops[iop + 1]]
            ).squeeze()
        return y

    def _rmatvec_serial(self, x):
        ncp = get_array_module(x)
        y = ncp.zeros(self.mops, dtype=self.dtype)
        for iop, oper in enumerate(self.ops):
            y[self.mmops[iop] : self.mmops[iop + 1]] = oper.rmatvec(
                x[self.nnops[iop] : self.nnops[iop + 1]]
            ).squeeze()
        return y

    def _matvec_multiproc(self, x):
        ys = self.pool.starmap(
            _matvec_rmatvec_map,
            [
                (oper._matvec, x[self.mmops[iop] : self.mmops[iop + 1]])
                for iop, oper in enumerate(self.ops)
            ],
        )
        y = np.hstack(ys)
        return y

    def _rmatvec_multiproc(self, x):
        ys = self.pool.starmap(
            _matvec_rmatvec_map,
            [
                (oper._rmatvec, x[self.nnops[iop] : self.nnops[iop + 1]])
                for iop, oper in enumerate(self.ops)
            ],
        )
        y = np.hstack(ys)
        return y

    def _matvec(self, x):
        if self.nproc == 1:
            y = self._matvec_serial(x)
        else:
            y = self._matvec_multiproc(x)
        return y

    def _rmatvec(self, x):
        if self.nproc == 1:
            y = self._rmatvec_serial(x)
        else:
            y = self._rmatvec_multiproc(x)
        return y



import numpy as np

from pylops import LinearOperator
from pylops.utils.backend import get_array_module


class Real(LinearOperator):
    r"""Real operator.

    Return the real component of the input. The adjoint returns a complex
    number with the same real component as the input and zero imaginary
    component.

    Parameters
    ----------
    dims : :obj:`int` or :obj:`tuple`
        Number of samples for each dimension
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

    Notes
    -----
    In forward mode:

    .. math::

        y_{i} = \Re\{x_{i}\} \quad \forall i=0,\ldots,N-1

    In adjoint mode:

    .. math::

        x_{i} = \Re\{y_{i}\} + 0i \quad \forall i=0,\ldots,N-1

    """

    def __init__(self, dims, dtype="complex128"):
        self.shape = (np.prod(np.array(dims)), np.prod(np.array(dims)))
        self.dtype = np.dtype(dtype)
        self.rdtype = np.real(np.ones(1, self.dtype)).dtype
        self.explicit = False
        self.clinear = False

    def _matvec(self, x):
        ncp = get_array_module(x)
        return ncp.real(x).astype(self.rdtype)

    def _rmatvec(self, x):
        ncp = get_array_module(x)
        return (ncp.real(x) + 0j).astype(self.dtype)



from itertools import chain

import numpy

from pylops.utils.backend import get_module


def _sampler_gaussian(m, batch_size, backend_module=numpy):
    return backend_module.random.randn(m, batch_size)


def _sampler_rayleigh(m, batch_size, backend_module=numpy):
    z = backend_module.random.randn(m, batch_size)
    for i in range(batch_size):
        z[:, i] *= m / backend_module.dot(z[:, i].T, z[:, i])
    return z


def _sampler_rademacher(m, batch_size, backend_module=numpy):
    return 2 * backend_module.random.binomial(1, 0.5, size=(m, batch_size)) - 1


_SAMPLERS = {
    "gaussian": _sampler_gaussian,
    "rayleigh": _sampler_rayleigh,
    "rademacher": _sampler_rademacher,
}


def trace_hutchinson(
    Op, neval=None, batch_size=None, sampler="rademacher", backend="numpy"
):
    r"""Trace of linear operator using the Hutchinson method.

    Returns an estimate of the trace of a linear operator using the Hutchinson
    method [1]_.

    Parameters
    ----------
    neval : :obj:`int`, optional
        Maximum number of matrix-vector products compute. Defaults to 10%
        of ``shape[1]``.
    batch_size : :obj:`int`, optional
        Vectorize computations by sampling sketching matrices instead of
        vectors. Set this value to as high as permitted by memory, but there is
        no guarantee of speedup. Coerced to never exceed ``neval``. When using
        "unitvector" as sampler, is coerced to not exceed ``shape[1]``.
        Defaults to 100 or ``neval``.
    sampler : :obj:`str`, optional
        Sample sketching matrices from the following distributions:

            - "gaussian": Mean zero, unit variance Gaussian.
            - "rayleigh": Sample from mean zero, unit variance Gaussian and
              normalize the columns.
            - "rademacher": Random sign.
            - "unitvector": Samples from the unit vectors :math:`\mathrm{e}_i`
              without replacement.

    backend : :obj:`str`, optional
        Backend used to densify matrix (``numpy`` or ``cupy``). Note that
        this must be consistent with how the operator has been created.

    Returns
    -------
    trace : :obj:`self.dtype`
        Operator trace.

    Raises
    -------
    ValueError
        If ``neval`` is smaller than 3.

    NotImplementedError
        If the ``sampler`` is not one of the available samplers.

    Notes
    -----
    Let :math:`m` = ``shape[1]`` and :math:`k` = ``neval``. This algorithm
    estimates the trace via

    .. math::
        \frac{1}{k}\sum\limits_{i=1}^k \mathbf{z}_i^T\,\mathbf{Op}\,\mathbf{z}_i

    where vectors :math:`\mathbf{z}_i` are sampled according to the sampling
    function. See [2]_ for a description of the variance and
    :math:`\epsilon`-approximation of different samplers.

    Prefer the Rademacher sampler if the goal is to minimize variance, but the
    Gaussian for a better probability of approximating the correct value. Use
    the Unit Vector approach if you are sampling a large number of ``neval``
    (compared to ``shape[1]``), especially if the operator is highly-structured.

    .. [1] Hutchinson, M. F. (1990). *A stochastic estimator of the trace of
           the influence matrix for laplacian smoothing splines*.
           Communications in Statistics - Simulation and Computation, 19(2),
           433–450.
    .. [2] Avron, H., and Toledo, S. (2011). *Randomized algorithms for
           estimating the trace of an implicit symmetric positive semi-definite
           matrix*. Journal of the ACM, 58(2), 1–34.
    """
    ncp = get_module(backend)
    m = Op.shape[1]
    neval = int(numpy.round(m * 0.1)) if neval is None else neval
    batch_size = min(neval, 100 if batch_size is None else batch_size)

    n_missing = neval - batch_size * (neval // batch_size)
    batch_range = chain(
        (batch_size for _ in range(0, neval - n_missing, batch_size)),
        (n_missing for _ in range(int(n_missing != 0))),
    )

    trace = ncp.zeros(1, dtype=Op.dtype)

    if sampler == "unitvector":
        remaining_vectors = list(range(m))
        n_total = 0
        while remaining_vectors:
            batch = min(batch_size, len(remaining_vectors))
            z = ncp.zeros((m, batch), dtype=Op.dtype)
            z_idx = ncp.random.choice(remaining_vectors, batch, replace=False)
            for i, idx in enumerate(z_idx):
                z[idx, i] = 1.0
                remaining_vectors.remove(idx)
            trace += ncp.trace((z.T @ (Op @ z)))
            n_total += batch
        trace *= m / n_total
        return trace[0]

    if sampler not in _SAMPLERS:
        raise NotImplementedError(f"sampler {sampler} not available.")

    sampler_fun = _SAMPLERS[sampler]
    for batch in batch_range:
        z = sampler_fun(m, batch, backend_module=ncp).astype(Op.dtype)
        trace += ncp.trace((z.T @ (Op @ z)))
    trace /= neval
    return trace[0]


def trace_hutchpp(Op, neval=None, sampler="rademacher", backend="numpy"):
    r"""Trace of linear operator using the Hutch++ method.

    Returns an estimate of the trace of a linear operator using the Hutch++
    method [1]_.

    Parameters
    ----------
    neval : :obj:`int`, optional
        Maximum number of matrix-vector products compute. Defaults to 10%
        of ``shape[1]``.
    sampler : :obj:`str`, optional
        Sample sketching matrices from the following distributions:

            - "gaussian": Mean zero, unit variance Gaussian.
            - "rayleigh": Sample from mean zero, unit variance Gaussian and
              normalize the columns.
            - "rademacher": Random sign.

    backend : :obj:`str`, optional
        Backend used to densify matrix (``numpy`` or ``cupy``). Note that
        this must be consistent with how the operator has been created.

    Returns
    -------
    trace : :obj:`self.dtype`
        Operator trace.

    Raises
    -------
    ValueError
        If ``neval`` is smaller than 3.

    NotImplementedError
        If the ``sampler`` is not one of the available samplers.

    Notes
    -----
    This function follows Algorithm 1 of [1]_. Let :math:`m` = ``shape[1]``
    and :math:`k` = ``neval``.

        1. Sample sketching matrices
           :math:`\mathbf{S} \in \mathbb{R}^{m \times \lfloor k/3\rfloor}`,
           and
           :math:`\mathbf{G} \in \mathbb{R}^{m \times \lfloor k/3\rfloor}`,
           from sub-Gaussian distributions.
        2. Compute reduced QR decomposition of :math:`\mathbf{Op}\,\mathbf{S}`,
           retaining only :math:`\mathbf{Q}`.
        3. Return :math:`\operatorname{tr}(\mathbf{Q}^T\,\mathbf{Op}\,\mathbf{Q}) + \frac{1}{\lfloor k/3\rfloor}\operatorname{tr}\left(\mathbf{G}^T(\mathbf{I} - \mathbf{Q}\mathbf{Q}^T)\,\mathbf{Op}\,(\mathbf{I} - \mathbf{Q}\mathbf{Q}^T)\mathbf{G}\right)`

    Use the Rademacher sampler unless you know what you are doing.

    .. [1] Meyer, R. A., Musco, C., Musco, C., & Woodruff, D. P. (2021).
        *Hutch++: Optimal Stochastic Trace Estimation*. In Symposium on Simplicity
        in Algorithms (SOSA) (pp. 142–155). Philadelphia, PA: Society for
        Industrial and Applied Mathematics. `link <https://arxiv.org/abs/2010.09649>`_
    """

    ncp = get_module(backend)
    m = Op.shape[1]

    neval = int(numpy.round(m * 0.1)) if neval is None else neval

    if sampler not in _SAMPLERS:
        raise NotImplementedError(f"sampler {sampler} not available.")

    sampler_fun = _SAMPLERS[sampler]

    batch = neval // 3
    if batch <= 0:
        msg = f"Sampler '{sampler}' not supported with {neval} samples."
        msg += " Try increasing it."
        raise ValueError(msg)

    S = sampler_fun(m, batch, backend_module=ncp).astype(Op.dtype)
    G = sampler_fun(m, batch, backend_module=ncp).astype(Op.dtype)

    Q, _ = ncp.linalg.qr(Op @ S)
    del S
    G = G - Q @ (Q.T @ G)

    trace = ncp.zeros(1, dtype=Op.dtype)
    trace += ncp.trace(Q.T @ (Op @ Q)) + ncp.trace(G.T @ (Op @ G)) / batch
    return trace[0]


def trace_nahutchpp(
    Op,
    neval=None,
    sampler="rademacher",
    c1=1.0 / 6.0,
    c2=1.0 / 3.0,
    backend="numpy",
):
    r"""Trace of linear operator using the NA-Hutch++ method.

    Returns an estimate of the trace of a linear operator using the
    Non-Adaptive variant of Hutch++ method [1]_.

    Parameters
    ----------
    neval : :obj:`int`, optional
        Maximum number of matrix-vector products compute. Defaults to 10%
        of ``shape[1]``.
    sampler : :obj:`str`, optional
        Sample sketching matrices from the following distributions:

            - "gaussian": Mean zero, unit variance Gaussian.
            - "rayleigh": Sample from mean zero, unit variance Gaussian and
              normalize the columns.
            - "rademacher": Random sign.

    c1 : :obj:`float`, optional
        Fraction of ``neval`` for sketching matrix :math:`\mathbf{S}`.
    c2 : :obj:`float`, optional
        Fraction of ``neval`` for sketching matrix :math:`\mathbf{R}`. Must be
        larger than ``c2``, ideally by a factor of at least 2.
    backend : :obj:`str`, optional
        Backend used to densify matrix (``numpy`` or ``cupy``). Note that
        this must be consistent with how the operator has been created.

    Returns
    -------
    trace : :obj:`self.dtype`
        Operator trace.

    Raises
    -------
    ValueError
        If ``neval`` not large enough to accomodate ``c1`` and ``c2``.

    NotImplementedError
        If the ``sampler`` is not one of the available samplers.

    Notes
    -----
    This function follows Algorithm 2 of [1]_. Let :math:`m` = ``shape[1]``
    and :math:`k` = ``neval``.

        1. Fix constants :math:`c_1`, :math:`c_2`, :math:`c_3` such that
           :math:`c_1 < c_2` and :math:`c_1 + c_2 + c_3 = 1`.
        2. Sample sketching matrices
           :math:`\mathbf{S} \in \mathbb{R}^{m \times c_1 k}`,
           :math:`\mathbf{R} \in \mathbb{R}^{m \times c_2 k}`,
           and
           :math:`\mathbf{G} \in \mathbb{R}^{m \times c_3 k}`
           from sub-Gaussian distributions.
        3. Compute :math:`\mathbf{Z} = \mathbf{Op}\,\mathbf{R}`,
           :math:`\mathbf{W} = \mathbf{Op}\,\mathbf{S}`, and
           :math:`\mathbf{Y} = (\mathbf{S}^T \mathbf{Z})^+`, where :math:`+`
           denotes the Moore–Penrose inverse.
        4. Return :math:`\operatorname{tr}(\mathbf{Y} \mathbf{W}^T \mathbf{Z}) + \frac{1}{c_3 k} \left[ \operatorname{tr}(\mathbf{G}^T\,\mathbf{Op}\,\mathbf{G}) - \operatorname{tr}(\mathbf{G}^T\mathbf{Z}\mathbf{Y}\mathbf{W}^T\mathbf{G})\right]`

    The default values for :math:`c_1` and :math:`c_2` are set to :math:`1/6`
    and :math:`1/3`, respectively, but [1]_ suggests :math:`1/4` and :math:`1/2`.

    Use the Rademacher sampler unless you know what you are doing.

    .. [1] Meyer, R. A., Musco, C., Musco, C., & Woodruff, D. P. (2021).
        *Hutch++: Optimal Stochastic Trace Estimation*. In Symposium on Simplicity
        in Algorithms (SOSA) (pp. 142–155). Philadelphia, PA: Society for
        Industrial and Applied Mathematics. `link <https://arxiv.org/abs/2010.09649>`_
    """

    ncp = get_module(backend)
    m = Op.shape[1]
    neval = int(numpy.round(m * 0.1)) if neval is None else neval

    if sampler not in _SAMPLERS:
        raise NotImplementedError(f"sampler {sampler} not available.")

    sampler_fun = _SAMPLERS[sampler]

    batch1 = int(numpy.round(neval * c1))
    batch2 = int(numpy.round(neval * c2))
    batch3 = neval - batch1 - batch2
    if batch1 <= 0 or batch2 <= 0 or batch3 <= 0:
        msg = f"Sampler '{sampler}' not supported with {neval} samples."
        msg += " Try increasing it."
        raise ValueError(msg)

    S = sampler_fun(m, batch1, backend_module=ncp).astype(Op.dtype)
    R = sampler_fun(m, batch2, backend_module=ncp).astype(Op.dtype)
    G = sampler_fun(m, batch3, backend_module=ncp).astype(Op.dtype)

    Z = Op @ R
    Wt = (Op @ S).T
    Y = ncp.linalg.pinv(S.T @ Z)
    trace = ncp.zeros(1, dtype=Op.dtype)
    trace += (
        ncp.trace(Y @ Wt @ Z)
        + (ncp.trace(G.T @ (Op @ G)) - ncp.trace(G.T @ Z @ Y @ Wt @ G)) / batch3
    )
    return trace[0]



from pylops.basicoperators import HStack, VStack


def _Block(
    ops, dtype=None, _HStack=HStack, _VStack=VStack, args_HStack={}, args_VStack={}
):
    """Block operator.

    Used to be able to provide operators from different libraries to
    Block.
    """
    hblocks = [_HStack(hblock, dtype=dtype, **args_HStack) for hblock in ops]
    return _VStack(hblocks, dtype=dtype, **args_VStack)


def Block(ops, nproc=1, dtype=None):
    r"""Block operator.

    Create a block operator from N lists of M linear operators each.

    Parameters
    ----------
    ops : :obj:`list`
        List of lists of operators to be combined in block fashion.
        Alternatively, :obj:`numpy.ndarray` or :obj:`scipy.sparse` matrices
        can be passed in place of one or more operators.
    nproc : :obj:`int`, optional
        Number of processes used to evaluate the N operators in parallel using
        ``multiprocessing``. If ``nproc=1``, work in serial mode.
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

    Notes
    -----
    In mathematics, a block or a partitioned matrix is a matrix that is
    interpreted as being broken into sections called blocks or submatrices.
    Similarly a block operator is composed of N sets of M linear operators
    each such that its application in forward mode leads to

    .. math::
        \begin{bmatrix}
            \mathbf{L}_{1,1}  & \mathbf{L}_{1,2} &  \ldots & \mathbf{L}_{1,M}  \\
            \mathbf{L}_{2,1}  & \mathbf{L}_{2,2} &  \ldots & \mathbf{L}_{2,M}  \\
            \vdots            & \vdots           &  \ddots & \vdots            \\
            \mathbf{L}_{N,1}  & \mathbf{L}_{N,2} &  \ldots & \mathbf{L}_{N,M}
        \end{bmatrix}
        \begin{bmatrix}
            \mathbf{x}_{1}  \\
            \mathbf{x}_{2}  \\
            \vdots          \\
            \mathbf{x}_{M}
        \end{bmatrix} =
        \begin{bmatrix}
            \mathbf{L}_{1,1} \mathbf{x}_{1} + \mathbf{L}_{1,2} \mathbf{x}_{2} +
            \mathbf{L}_{1,M} \mathbf{x}_{M} \\
            \mathbf{L}_{2,1} \mathbf{x}_{1} + \mathbf{L}_{2,2} \mathbf{x}_{2} +
            \mathbf{L}_{2,M} \mathbf{x}_{M} \\
            \vdots     \\
            \mathbf{L}_{N,1} \mathbf{x}_{1} + \mathbf{L}_{N,2} \mathbf{x}_{2} +
            \mathbf{L}_{N,M} \mathbf{x}_{M}
        \end{bmatrix}

    while its application in adjoint mode leads to

    .. math::
        \begin{bmatrix}
            \mathbf{L}_{1,1}^H  & \mathbf{L}_{2,1}^H & \ldots & \mathbf{L}_{N,1}^H  \\
            \mathbf{L}_{1,2}^H  & \mathbf{L}_{2,2}^H & \ldots & \mathbf{L}_{N,2}^H  \\
            \vdots              & \vdots             & \ddots & \vdots \\
            \mathbf{L}_{1,M}^H  & \mathbf{L}_{2,M}^H & \ldots & \mathbf{L}_{N,M}^H
        \end{bmatrix}
        \begin{bmatrix}
            \mathbf{y}_{1}  \\
            \mathbf{y}_{2}  \\
            \vdots     \\
            \mathbf{y}_{N}
        \end{bmatrix} =
        \begin{bmatrix}
            \mathbf{L}_{1,1}^H \mathbf{y}_{1} +
            \mathbf{L}_{2,1}^H \mathbf{y}_{2} +
            \mathbf{L}_{N,1}^H \mathbf{y}_{N} \\
            \mathbf{L}_{1,2}^H \mathbf{y}_{1} +
            \mathbf{L}_{2,2}^H \mathbf{y}_{2} +
            \mathbf{L}_{N,2}^H \mathbf{y}_{N} \\
            \vdots     \\
            \mathbf{L}_{1,M}^H \mathbf{y}_{1} +
            \mathbf{L}_{2,M}^H \mathbf{y}_{2} +
            \mathbf{L}_{N,M}^H \mathbf{y}_{N}
        \end{bmatrix}

    """
    return _Block(ops, dtype=dtype, args_VStack={"nproc": nproc})



import logging
import warnings
from enum import Enum, auto

import numpy as np
from numpy.core.multiarray import normalize_axis_index

from pylops import LinearOperator
from pylops.utils.backend import get_complex_dtype, get_real_dtype

logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.WARNING)


class _FFTNorms(Enum):
    ORTHO = auto()
    NONE = auto()
    ONE_OVER_N = auto()


def _value_or_list_like_to_array(value_or_list_like, repeat=1):
    """Convert an object which is either single value or a list-like to an array.

    Parameters
    ----------
    value_or_list_like
        Single value or list-like.
    repeat : `obj`:`int`
        Size of resulting array if value is passed. If list is passed, it is ignored.

    Returns
    -------
    out : `obj`:`numpy.array`
        When the input is a single value, returned an array with `repeat` samples
        containing that value. When the input is a list-like object, converts it to an
        array.

    """
    try:
        len(value_or_list_like)
        out = np.array(value_or_list_like)
    except TypeError:
        out = np.array([value_or_list_like] * repeat)
    return out


def _raise_on_wrong_dtype(arr, dtype, name):
    """Raises an error if dtype of `arr` is not a subdtype of `dtype`.

    Parameters
    ----------
    arr : `obj`:`numpy.array`
        Array whose type will be checked
    dtype : `obj`:`numpy.dtype`
        Type which must be a supertype of `arr.dtype`.
    name : `obj`:`str`
        Name of parameter to issue error.

    Raises
    ------
    TypeError
        When `arr.dtype` is not a subdtype of `dtype`.

    """
    if not np.issubdtype(arr.dtype, dtype):
        raise TypeError(
            (
                f"Wrong input type for `{name}`. "
                f'Must be "{dtype}", but received to "{arr.dtype}".'
            )
        )


class _BaseFFT(LinearOperator):
    """Base class for one dimensional Fast-Fourier Transform"""

    def __init__(
        self,
        dims,
        dir=0,
        nfft=None,
        sampling=1.0,
        norm="ortho",
        real=False,
        ifftshift_before=False,
        fftshift_after=False,
        dtype="complex128",
    ):
        self.dims = _value_or_list_like_to_array(dims)
        _raise_on_wrong_dtype(self.dims, np.integer, "dims")

        self.ndim = len(self.dims)

        dirs = _value_or_list_like_to_array(dir)
        _raise_on_wrong_dtype(dirs, np.integer, "dir")
        self.dir = normalize_axis_index(dirs[0], self.ndim)

        if nfft is None:
            nfft = self.dims[self.dir]
        else:
            nffts = _value_or_list_like_to_array(nfft)
            _raise_on_wrong_dtype(nffts, np.integer, "nfft")
            nfft = nffts[0]
        self.nfft = nfft

        # Check if the user provided nfft smaller than n (size of signal in
        # original domain). If so, raise a warning as this is unlikely a
        # wanted behavoir (since FFT routines cut some of the input signal
        # before applying fft, which is lost forever) and set a flag such that
        # a padding is applied after ifft
        self.doifftpad = False
        if self.nfft < self.dims[dir]:
            self.doifftpad = True
            self.ifftpad = [(0, 0)] * self.ndim
            self.ifftpad[self.dir] = (0, self.dims[dir] - self.nfft)
            warnings.warn(
                f"nfft={self.nfft} has been selected to be smaller than the size of the original signal (n=self.dims[dir]). "
                f"This is rarely intended behavior as the original signal will be truncated prior to applying fft, "
                f"if this is the required behaviour ignore this message."
            )

        if norm == "ortho":
            self.norm = _FFTNorms.ORTHO
        elif norm == "none":
            self.norm = _FFTNorms.NONE
        elif norm.lower() == "1/n":
            self.norm = _FFTNorms.ONE_OVER_N
        elif norm == "backward":
            raise ValueError(
                'To use no scaling on the forward transform, use "none". Note that in this case, the adjoint transform will *not* have a 1/n scaling.'
            )
        elif norm == "forward":
            raise ValueError(
                'To use 1/n scaling on the forward transform, use "1/n". Note that in this case, the adjoint transform will *also* have a 1/n scaling.'
            )
        else:
            raise ValueError(f"'{norm}' is not one of 'ortho', 'none' or '1/n'")

        self.real = real
        self.ifftshift_before = ifftshift_before

        self.f = (
            np.fft.rfftfreq(self.nfft, d=sampling)
            if real
            else np.fft.fftfreq(self.nfft, d=sampling)
        )
        self.fftshift_after = fftshift_after
        if self.fftshift_after:
            if self.real:
                warnings.warn(
                    "Using fftshift_after with real=True. fftshift should only be applied after a complex FFT. This is rarely intended behavior but if it is, ignore this message."
                )
            self.f = np.fft.fftshift(self.f)

        self.dims_fft = self.dims.copy()
        self.dims_fft[self.dir] = self.nfft // 2 + 1 if self.real else self.nfft
        self.shape = (int(np.prod(self.dims_fft)), int(np.prod(self.dims)))

        # Find types to enforce to forward and adjoint outputs. This is
        # required as np.fft.fft always returns complex128 even if input is
        # float32 or less. Moreover, when choosing real=True, the type of the
        # adjoint output is forced to be real even if the provided dtype
        # is complex.
        self.rdtype = get_real_dtype(dtype) if self.real else np.dtype(dtype)
        self.cdtype = get_complex_dtype(dtype)
        self.dtype = self.cdtype
        self.clinear = False if self.real or np.issubdtype(dtype, np.floating) else True
        self.explicit = False

    def _matvec(self, x):
        raise NotImplementedError(
            "_BaseFFT does not provide _matvec. It must be implemented separately."
        )

    def _rmatvec(self, x):
        raise NotImplementedError(
            "_BaseFFT does not provide _rmatvec. It must be implemented separately."
        )


class _BaseFFTND(LinearOperator):
    """Base class for N-dimensional fast Fourier Transform"""

    def __init__(
        self,
        dims,
        dirs=None,
        nffts=None,
        sampling=1.0,
        norm="ortho",
        real=False,
        ifftshift_before=False,
        fftshift_after=False,
        dtype="complex128",
    ):
        self.dims = _value_or_list_like_to_array(dims)
        _raise_on_wrong_dtype(self.dims, np.integer, "dims")

        self.ndim = len(self.dims)

        dirs = _value_or_list_like_to_array(dirs)
        _raise_on_wrong_dtype(dirs, np.integer, "dirs")
        self.dirs = np.array([normalize_axis_index(d, self.ndim) for d in dirs])
        self.ndirs = len(self.dirs)
        if self.ndirs != len(np.unique(self.dirs)):
            warnings.warn(
                "At least one direction is repeated. This may cause unexpected results."
            )

        nffts = _value_or_list_like_to_array(nffts, repeat=self.ndirs)
        if len(nffts[np.equal(nffts, None)]) > 0:  # Found None(s) in nffts
            nffts[np.equal(nffts, None)] = np.array(
                [self.dims[d] for d, n in zip(dirs, nffts) if n is None]
            )
            nffts = nffts.astype(self.dims.dtype)
        self.nffts = nffts
        _raise_on_wrong_dtype(self.nffts, np.integer, "nffts")

        sampling = _value_or_list_like_to_array(sampling, repeat=self.ndirs)
        if np.issubdtype(sampling.dtype, np.integer):  # Promote to float64 if integer
            sampling = sampling.astype(np.float64)
        self.sampling = sampling
        _raise_on_wrong_dtype(self.sampling, np.floating, "sampling")

        self.ifftshift_before = _value_or_list_like_to_array(
            ifftshift_before, repeat=self.ndirs
        )
        _raise_on_wrong_dtype(self.ifftshift_before, bool, "ifftshift_before")

        self.fftshift_after = _value_or_list_like_to_array(
            fftshift_after, repeat=self.ndirs
        )
        _raise_on_wrong_dtype(self.fftshift_after, bool, "fftshift_after")

        if (
            self.ndirs != len(self.nffts)
            or self.ndirs != len(self.sampling)
            or self.ndirs != len(self.ifftshift_before)
            or self.ndirs != len(self.fftshift_after)
        ):
            raise ValueError(
                (
                    "`dirs`, `nffts`, `sampling`, `ifftshift_before` and "
                    "`fftshift_after` must the have same number of elements. Received "
                    f"{self.ndirs}, {len(self.nffts)}, {len(self.sampling)}, "
                    f"{len(self.ifftshift_before)} and {len(self.fftshift_after)}, "
                    "respectively."
                )
            )

        # Check if the user provided nfft smaller than n. See _BaseFFT for
        # details
        nfftshort = [
            nfft < self.dims[direction]
            for direction, nfft in zip(self.dirs, self.nffts)
        ]
        self.doifftpad = any(nfftshort)
        if self.doifftpad:
            self.ifftpad = [(0, 0)] * self.ndim
            for idir, (direction, nfshort) in enumerate(zip(self.dirs, nfftshort)):
                if nfshort:
                    self.ifftpad[direction] = (
                        0,
                        self.dims[direction] - self.nffts[idir],
                    )
            warnings.warn(
                f"nffts in directions {np.where(nfftshort)[0]} have been selected to be smaller than the size of the original signal. "
                f"This is rarely intended behavior as the original signal will be truncated prior to applying fft, "
                f"if this is the required behaviour ignore this message."
            )

        if norm == "ortho":
            self.norm = _FFTNorms.ORTHO
        elif norm == "none":
            self.norm = _FFTNorms.NONE
        elif norm.lower() == "1/n":
            self.norm = _FFTNorms.ONE_OVER_N
        elif norm == "backward":
            raise ValueError(
                'To use no scaling on the forward transform, use "none". Note that in this case, the adjoint transform will *not* have a 1/n scaling.'
            )
        elif norm == "forward":
            raise ValueError(
                'To use 1/n scaling on the forward transform, use "1/n". Note that in this case, the adjoint transform will *also* have a 1/n scaling.'
            )
        else:
            raise ValueError(f"'{norm}' is not one of 'ortho', 'none' or '1/n'")

        self.real = real

        fs = [
            np.fft.fftshift(np.fft.fftfreq(n, d=s))
            if fftshift
            else np.fft.fftfreq(n, d=s)
            for n, s, fftshift in zip(self.nffts, self.sampling, self.fftshift_after)
        ]
        if self.real:
            fs[-1] = np.fft.rfftfreq(self.nffts[-1], d=self.sampling[-1])
            if self.fftshift_after[-1]:
                warnings.warn(
                    (
                        "Using real=True and fftshift_after on the last direction. "
                        "fftshift should only be applied on directions with negative "
                        "and positive frequencies. When using FFTND with real=True, "
                        "are all directions except the last. If you wish to proceed "
                        "applying fftshift on a frequency axis with only positive "
                        "frequencies, ignore this message."
                    )
                )
                fs[-1] = np.fft.fftshift(fs[-1])
        self.fs = tuple(fs)
        self.dims_fft = self.dims.copy()
        self.dims_fft[self.dirs] = self.nffts
        if self.real:
            self.dims_fft[self.dirs[-1]] = self.nffts[-1] // 2 + 1
        self.shape = (int(np.prod(self.dims_fft)), int(np.prod(self.dims)))
        self.rdtype = get_real_dtype(dtype) if self.real else np.dtype(dtype)
        self.cdtype = get_complex_dtype(dtype)
        self.dtype = self.cdtype
        self.clinear = False if self.real or np.issubdtype(dtype, np.floating) else True
        self.explicit = False

    def _matvec(self, x):
        raise NotImplementedError(
            "_BaseFFT does not provide _matvec. It must be implemented separately."
        )

    def _rmatvec(self, x):
        raise NotImplementedError(
            "_BaseFFT does not provide _rmatvec. It must be implemented separately."
        )



import numpy as np

from pylops import LinearOperator
from pylops.utils.backend import get_array_module


class Fredholm1(LinearOperator):
    r"""Fredholm integral of first kind.

    Implement a multi-dimensional Fredholm integral of first kind. Note that if
    the integral is two dimensional, this can be directly implemented using
    :class:`pylops.basicoperators.MatrixMult`. A multi-dimensional
    Fredholm integral can be performed as a :class:`pylops.basicoperators.BlockDiag`
    operator of a series of :class:`pylops.basicoperators.MatrixMult`. However,
    here we take advantage of the structure of the kernel and perform it in a
    more efficient manner.

    Parameters
    ----------
    G : :obj:`numpy.ndarray`
        Multi-dimensional convolution kernel of size
        :math:`[n_{\text{slice}} \times n_x \times n_y]`
    nz : :obj:`numpy.ndarray`, optional
        Additional dimension of model
    saveGt : :obj:`bool`, optional
        Save ``G`` and ``G.H`` to speed up the computation of adjoint
        (``True``) or create ``G.H`` on-the-fly (``False``)
        Note that ``saveGt=True`` will double the amount of required memory
    usematmul : :obj:`bool`, optional
        Use :func:`numpy.matmul` (``True``) or for-loop with :func:`numpy.dot`
        (``False``). As it is not possible to define which approach is more
        performant (this is highly dependent on the size of ``G`` and input
        arrays as well as the hardware used in the compution), we advise users
        to time both methods for their specific problem prior to making a
        choice.
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved
        explicitly (``True``) or not (``False``)

    Notes
    -----
    A multi-dimensional Fredholm integral of first kind can be expressed as

    .. math::

        d(k, x, z) = \int{G(k, x, y) m(k, y, z) \,\mathrm{d}y}
        \quad \forall k=1,\ldots,n_{slice}

    on the other hand its adjoin is expressed as

    .. math::

        m(k, y, z) = \int{G^*(k, y, x) d(k, x, z) \,\mathrm{d}x}
        \quad \forall k=1,\ldots,n_{\text{slice}}

    In discrete form, this operator can be seen as a block-diagonal
    matrix multiplication:

    .. math::
        \begin{bmatrix}
            \mathbf{G}_{k=1}  & \mathbf{0}       &  \ldots & \mathbf{0} \\
            \mathbf{0}        & \mathbf{G}_{k=2} &  \ldots & \mathbf{0} \\
            \vdots            & \vdots           &  \ddots & \vdots        \\
            \mathbf{0}        & \mathbf{0}       &  \ldots & \mathbf{G}_{k=N}
        \end{bmatrix}
        \begin{bmatrix}
            \mathbf{m}_{k=1}  \\
            \mathbf{m}_{k=2}  \\
            \vdots            \\
            \mathbf{m}_{k=N}
        \end{bmatrix}

    """

    def __init__(self, G, nz=1, saveGt=True, usematmul=True, dtype="float64"):
        self.nz = nz
        self.nsl, self.nx, self.ny = G.shape
        self.G = G
        if saveGt:
            self.GT = G.transpose((0, 2, 1)).conj()
        self.usematmul = usematmul
        self.shape = (self.nsl * self.nx * self.nz, self.nsl * self.ny * self.nz)
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        ncp = get_array_module(x)
        x = ncp.squeeze(x.reshape(self.nsl, self.ny, self.nz))
        if self.usematmul:
            if self.nz == 1:
                x = x[..., ncp.newaxis]
            y = ncp.matmul(self.G, x)
        else:
            y = ncp.squeeze(ncp.zeros((self.nsl, self.nx, self.nz), dtype=self.dtype))
            for isl in range(self.nsl):
                y[isl] = ncp.dot(self.G[isl], x[isl])
        return y.ravel()

    def _rmatvec(self, x):
        ncp = get_array_module(x)
        x = ncp.squeeze(x.reshape(self.nsl, self.nx, self.nz))
        if self.usematmul:
            if self.nz == 1:
                x = x[..., ncp.newaxis]
            if hasattr(self, "GT"):
                y = ncp.matmul(self.GT, x)
            else:
                # y = ncp.matmul(self.G.transpose((0, 2, 1)).conj(), x)
                y = (
                    ncp.matmul(x.transpose(0, 2, 1).conj(), self.G)
                    .transpose(0, 2, 1)
                    .conj()
                )
        else:
            y = ncp.squeeze(ncp.zeros((self.nsl, self.ny, self.nz), dtype=self.dtype))
            if hasattr(self, "GT"):
                for isl in range(self.nsl):
                    y[isl] = ncp.dot(self.GT[isl], x[isl])
            else:
                for isl in range(self.nsl):
                    # y[isl] = ncp.dot(self.G[isl].conj().T, x[isl])
                    y[isl] = ncp.dot(x[isl].T.conj(), self.G[isl]).T.conj()
        return y.ravel()



import logging

import numpy as np
import scipy as sp
from scipy.sparse.linalg import inv

from pylops import LinearOperator
from pylops.utils.backend import get_array_module

logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.WARNING)


class MatrixMult(LinearOperator):
    r"""Matrix multiplication.

    Simple wrapper to :py:func:`numpy.dot` and :py:func:`numpy.vdot` for
    an input matrix :math:`\mathbf{A}`.

    Parameters
    ----------
    A : :obj:`numpy.ndarray` or :obj:`scipy.sparse` matrix
        Matrix.
    dims : :obj:`tuple`, optional
        Number of samples for each other dimension of model
        (model/data will be reshaped and ``A`` applied multiple times
        to each column of the model/data).
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly
        (``True``) or not (``False``)
    complex : :obj:`bool`
        Matrix has complex numbers (``True``) or not (``False``)

    """

    def __init__(self, A, dims=None, dtype="float64"):
        ncp = get_array_module(A)
        self.A = A
        if isinstance(A, ncp.ndarray):
            self.complex = np.iscomplexobj(A)
        else:
            self.complex = np.iscomplexobj(A.data)
        if dims is None:
            self.reshape = False
            self.shape = A.shape
            self.explicit = True
        else:
            if isinstance(dims, int):
                dims = (dims,)
            self.reshape = True
            self.dims = np.array(dims, dtype=int)
            self.reshapedims = [
                np.insert([np.prod(self.dims)], 0, self.A.shape[1]),
                np.insert([np.prod(self.dims)], 0, self.A.shape[0]),
            ]
            self.shape = (
                A.shape[0] * np.prod(self.dims),
                A.shape[1] * np.prod(self.dims),
            )
            self.explicit = False
        self.dtype = np.dtype(dtype)
        # Check dtype for correctness (upcast to complex when A is complex)
        if np.iscomplexobj(A) and not np.iscomplexobj(np.ones(1, dtype=self.dtype)):
            self.dtype = A.dtype
            logging.warning(
                "Matrix A is a complex object, dtype cast to %s" % self.dtype
            )

    def _matvec(self, x):
        ncp = get_array_module(x)
        if self.reshape:
            x = ncp.reshape(x, self.reshapedims[0])
        y = self.A.dot(x)
        if self.reshape:
            return y.ravel()
        else:
            return y

    def _rmatvec(self, x):
        ncp = get_array_module(x)
        if self.reshape:
            x = ncp.reshape(x, self.reshapedims[1])
        if self.complex:
            y = (self.A.T.dot(x.conj())).conj()
        else:
            y = self.A.T.dot(x)

        if self.reshape:
            return y.ravel()
        else:
            return y

    def inv(self):
        r"""Return the inverse of :math:`\mathbf{A}`.

        Returns
        ----------
        Ainv : :obj:`numpy.ndarray`
            Inverse matrix.

        """
        if sp.sparse.issparse(self.A):
            Ainv = inv(self.A)
        else:
            ncp = get_array_module(self.A)
            Ainv = ncp.linalg.inv(self.A)
        return Ainv



import random
import string

from scipy.sparse.linalg.interface import (
    _AdjointLinearOperator,
    _ProductLinearOperator,
    _SumLinearOperator,
    _TransposedLinearOperator,
)

from pylops import LinearOperator
from pylops.basicoperators import BlockDiag, HStack, VStack
from pylops.LinearOperator import _ScaledLinearOperator

try:
    from sympy import BlockDiagMatrix, BlockMatrix, MatrixSymbol
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Sympy package not installed. In order to use "
        "the describe method run "
        "install sympy."
    )

compositeops = (
    LinearOperator,
    _SumLinearOperator,
    _ProductLinearOperator,
    _ScaledLinearOperator,
    _AdjointLinearOperator,
    _TransposedLinearOperator,
    HStack,
    VStack,
    BlockDiag,
)


def _in_notebook():
    """Check if code is running inside notebook"""
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def _assign_name(Op, Ops, names):
    """Assign name to an operator as provided by the user
    (or randomly select one when not provided by the user)

    Parameters
    ----------
    Op : :obj:`pylops.LinearOperator`
        Linear Operator to assign name to
    Ops : :obj:`dict`
        Dictionary of Operators found by the _describe method whilst crawling
        through the composite operator to describe
    names : :obj:`list`
        List of currently assigned names

    """
    # Add a suffix when all letters of the alphabet are already in use. This
    # decision is made by counting the length of the names list and using the
    # English vocabulary (26 characters)
    suffix = "1" * (len(names) // 26)
    # Propose a new name, where a random letter is chosen if Op does not
    # have a name
    proposedname = (
        getattr(Op, "name", random.choice(string.ascii_letters).upper()) + suffix
    )
    if proposedname not in names or (Ops[proposedname][1] == id(Op)):
        # Assign the proposed name if this is not yet in use or if it is
        # used by the same operator. Note that an operator may reapper
        # multiple times in an expression
        name = proposedname
    else:
        # Propose a new name until an unused character is found
        origname = proposedname
        while proposedname in names:
            proposedname = random.choice(string.ascii_letters).upper() + suffix
        name = proposedname
        print(
            f"The user has used the same name {origname} for two distinct operators, "
            f"changing name of operator {type(Op).__name__} to {name}..."
        )
    Op.name = name
    return name


def _describeop(Op, Ops, names):
    """Core steps to describe a single operator

    Parameters
    ----------
    Op : :obj:`pylops.LinearOperator`
        Linear Operator to assign name to
    Ops : :obj:`dict`
        Dictionary of Operators found by the _describe method whilst crawling
        through the composite operator to describe
    names : :obj:`list`
        List of currently assigned names

    """
    if type(Op) not in compositeops:
        # A native PyLops operator has been found, assign a name and store
        # it as MatrixSymbol
        name = _assign_name(Op, Ops, names)
        Op0 = MatrixSymbol(name, 1, 1)
        Ops_ = {name: (type(Op).__name__, id(Op))}
    elif type(Op) == LinearOperator:
        # A LinearOperator has been found, either extract Op and start to
        # describe it or if a name has been given to the operator treat as
        # it is (this is useful when users do not want an operator to be
        # further disected into its components
        name = getattr(Op, "name", None)
        if name is None:
            Op0, Ops_ = _describe(Op.Op)
        else:
            Ops_ = {name: (type(Op).__name__, id(Op))}
            Op0 = MatrixSymbol(name, 1, 1)
    else:
        # When finding a composite operator, send it again to the _describe
        # method
        Op0, Ops_ = _describe(Op)
    return Op0, Ops_


def _describe(Op):
    """Core steps to describe a composite operator. This is done recursively.

    Parameters
    ----------
    Op : :obj:`pylops.LinearOperator`
        Linear Operator to assign name to
    Ops : :obj:`dict`
        Dictionary of Operators found by the _describe method whilst crawling
        through the composite operator to describe
    names : :obj:`list`
        List of currently assigned names

    """
    Ops = {}
    if type(Op) not in compositeops:
        # A native PyLops operator has been found, assign a name and store
        # it as MatrixSymbol
        name = _assign_name(Op, Ops, list(Ops.keys()))
        Ops[name] = (type(Op).__name__, id(Op))
        Opsym = MatrixSymbol(Op.name, 1, 1)
    else:
        if type(Op) == LinearOperator:
            # A LinearOperator has been found, either extract Op and start to
            # describe it or if a name has been given to the operator treat as
            # it is (this is useful when users do not want an operator to be
            # further disected into its components
            name = getattr(Op, "name", None)
            if name is None:
                Opsym, Ops_ = _describe(Op.Op)
                Ops.update(Ops_)
            else:
                Ops[name] = (type(Op).__name__, id(Op))
                Opsym = MatrixSymbol(Op.name, 1, 1)
        elif type(Op) == _AdjointLinearOperator:
            # An adjoint LinearOperator has been found, describe it and attach
            # the adjoint symbol to its sympy representation
            Opsym, Ops_ = _describe(Op.args[0])
            Opsym = Opsym.adjoint()
            Ops.update(Ops_)
        elif type(Op) == _TransposedLinearOperator:
            # A transposed LinearOperator has been found, describe it and
            # attach the transposed symbol to its sympy representation
            Opsym, Ops_ = _describe(Op.args[0])
            Opsym = Opsym.T
            Ops.update(Ops_)
        elif type(Op) == _ScaledLinearOperator:
            # A scaled LinearOperator has been found, describe it and
            # attach the scaling to its sympy representation. Note that the
            # scaling could either on the left or right side of the operator,
            # so we need to try both
            if isinstance(Op.args[0], LinearOperator):
                Opsym, Ops_ = _describeop(Op.args[0], Ops, list(Ops.keys()))
                Opsym = Op.args[1] * Opsym
                Ops.update(Ops_)
            else:
                Opsym, Ops_ = _describeop(Op.args[1], Ops, list(Ops.keys()))
                Opsym = Op.args[1] * Opsym
                Ops.update(Ops_)
        elif type(Op) == _SumLinearOperator:
            # A sum LinearOperator has been found, describe both operators
            # either side of the plus sign and sum their sympy representations
            Opsym0, Ops_ = _describeop(Op.args[0], Ops, list(Ops.keys()))
            Ops.update(Ops_)
            Opsym1, Ops_ = _describeop(Op.args[1], Ops, list(Ops.keys()))
            Ops.update(Ops_)
            Opsym = Opsym0 + Opsym1
        elif type(Op) == _ProductLinearOperator:
            # Same as sum LinearOperator but for product
            Opsym0, Ops_ = _describeop(Op.args[0], Ops, list(Ops.keys()))
            Ops.update(Ops_)
            Opsym1, Ops_ = _describeop(Op.args[1], Ops, list(Ops.keys()))
            Ops.update(Ops_)
            Opsym = Opsym0 * Opsym1
        elif type(Op) in (VStack, HStack, BlockDiag):
            # A special composite operator has been found, stack its components
            # horizontally, vertically, or along a diagonal
            Opsyms = []
            for op in Op.ops:
                Opsym, Ops_ = _describeop(op, Ops, list(Ops.keys()))
                Opsyms.append(Opsym)
                Ops.update(Ops_)
            Ops.update(Ops_)
            if type(Op) == VStack:
                Opsym = BlockMatrix([[Opsym] for Opsym in Opsyms])
            elif type(Op) == HStack:
                Opsym = BlockMatrix(Opsyms)
            elif type(Op) == BlockDiag:
                Opsym = BlockDiagMatrix(*Opsyms)
    return Opsym, Ops


def describe(Op):
    """Describe a PyLops operator

    .. versionadded:: 1.17.0

    Convert a PyLops operator into a ``sympy`` mathematical formula.
    This routine is useful both for debugging and educational purposes.

    Note that users can add a name to each operator prior to
    running the describe method, i.e. ``Op.name='Opname'``. Alternatively, each
    of the PyLops operator that composes the operator ``Op`` is automatically
    assigned a name. Moreover, note that the symbols :math:`T` and
    :math:`\dagger` are used in the mathematical expressions to indicate
    transposed and adjoint operators, respectively.

    Parameters
    ----------
    Op : :obj:`pylops.LinearOperator`
        Linear Operator to describe

    """
    # Describe the operator
    Opsym, Ops = _describe(Op)
    # Clean up Ops from id
    Ops = {op: Ops[op][0] for op in Ops.keys()}
    # Check if this command is run in a Jupyter notebook or normal shell and
    # display the operator accordingly
    if _in_notebook():
        from IPython.display import display

        display(Opsym)
    else:
        print(Opsym)
    print("where:", Ops)



import numpy as np

from pylops.basicoperators import Diagonal
from pylops.signalprocessing import FFT


def Shift(
    dims,
    shift,
    dir=0,
    nfft=None,
    sampling=1.0,
    real=False,
    engine="numpy",
    dtype="complex128",
    **kwargs_fftw
):
    r"""Shift operator

    Apply fractional shift in the frequency domain along a specific direction
    ``dir`` of a multi-dimensional array of size ``dim``.

    Parameters
    ----------
    dims : :obj:`tuple`
        Number of samples for each dimension
    shift : :obj:`float`
        Fractional shift to apply in the same unit as ``sampling``.
    dir : :obj:`int`, optional
        Direction along which FFT is applied.
    nfft : :obj:`int`, optional
        Number of samples in Fourier Transform (same as input if ``nfft=None``)
    sampling : :obj:`float`, optional
        Sampling step :math:`\Delta t`.
    real : :obj:`bool`, optional
        Model to which fft is applied has real numbers (``True``) or not
        (``False``). Used to enforce that the output of adjoint of a real
        model is real.
    engine : :obj:`str`, optional
        Engine used for fft computation (``numpy``, ``scipy``, or ``fftw``). Choose
        ``numpy`` when working with CuPy arrays.
    dtype : :obj:`str`, optional
        Type of elements in input array.
    **kwargs_fftw
            Arbitrary keyword arguments
            for :py:class:`pyfftw.FTTW`

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly
        (``True``) or not (``False``)

    Raises
    ------
    ValueError
        If ``dims`` is provided and ``dir`` is bigger than ``len(dims)``
    NotImplementedError
        If ``engine`` is neither ``numpy``, ``scipy``, nor ``fftw``

    Notes
    -----
    The Shift operator applies the forward Fourier transform, an element-wise
    complex scaling, and inverse fourier transform

    .. math::
         \mathbf{y}= \mathbf{F}^{-1} \mathbf{S} \mathbf{F} \mathbf{x}

    Here :math:`\mathbf{S}` is a diagonal operator that scales the Fourier
    transformed input by :math:`e^{-j2\pi f t_S}`, where :math:`t_S` is the
    chosen ``shift``.

    """
    # TODO: Use offer the same keywords as new FFT
    Fop = FFT(
        dims, dir, nfft, sampling, real=real, engine=engine, dtype=dtype, **kwargs_fftw
    )
    if isinstance(dims, int):
        dimsdiag = None
    else:
        dimsdiag = list(dims)
        dimsdiag[dir] = len(Fop.f)
    shift = np.exp(-1j * 2 * np.pi * Fop.f * shift)
    Sop = Diagonal(shift, dims=dimsdiag, dir=dir, dtype=Fop.cdtype)
    Op = Fop.H * Sop * Fop
    # force dtype to that of input (FFT always upcasts it to complex)
    Op.dtype = dtype
    return Op



import logging
import warnings

import numpy as np
import scipy.fft

from pylops.signalprocessing._BaseFFTs import _BaseFFTND, _FFTNorms

logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.WARNING)


class _FFT2D_numpy(_BaseFFTND):
    """Two dimensional Fast-Fourier Transform using NumPy"""

    def __init__(
        self,
        dims,
        dirs=(0, 1),
        nffts=None,
        sampling=1.0,
        norm="ortho",
        real=False,
        ifftshift_before=False,
        fftshift_after=False,
        dtype="complex128",
    ):
        super().__init__(
            dims=dims,
            dirs=dirs,
            nffts=nffts,
            sampling=sampling,
            norm=norm,
            real=real,
            ifftshift_before=ifftshift_before,
            fftshift_after=fftshift_after,
            dtype=dtype,
        )
        if self.cdtype != np.complex128:
            warnings.warn(
                f"numpy backend always returns complex128 dtype. To respect the passed dtype, data will be casted to {self.cdtype}."
            )

        # checks
        if self.ndim < 2:
            raise ValueError("FFT2D requires at least two input dimensions")
        if self.ndirs != 2:
            raise ValueError("FFT2D must be applied along exactly two dimensions")

        self.f1, self.f2 = self.fs
        del self.fs

        self._norm_kwargs = {"norm": None}  # equivalent to "backward" in Numpy/Scipy
        if self.norm is _FFTNorms.ORTHO:
            self._norm_kwargs["norm"] = "ortho"
        elif self.norm is _FFTNorms.NONE:
            self._scale = np.prod(self.nffts)
        elif self.norm is _FFTNorms.ONE_OVER_N:
            self._scale = 1.0 / np.prod(self.nffts)

    def _matvec(self, x):
        x = np.reshape(x, self.dims)
        if self.ifftshift_before.any():
            x = np.fft.ifftshift(x, axes=self.dirs[self.ifftshift_before])
        if not self.clinear:
            x = np.real(x)
        if self.real:
            y = np.fft.rfft2(x, s=self.nffts, axes=self.dirs, **self._norm_kwargs)
            # Apply scaling to obtain a correct adjoint for this operator
            y = np.swapaxes(y, -1, self.dirs[-1])
            y[..., 1 : 1 + (self.nffts[-1] - 1) // 2] *= np.sqrt(2)
            y = np.swapaxes(y, self.dirs[-1], -1)
        else:
            y = np.fft.fft2(x, s=self.nffts, axes=self.dirs, **self._norm_kwargs)
        if self.norm is _FFTNorms.ONE_OVER_N:
            y *= self._scale
        y = y.astype(self.cdtype)
        if self.fftshift_after.any():
            y = np.fft.fftshift(y, axes=self.dirs[self.fftshift_after])
        return y.ravel()

    def _rmatvec(self, x):
        x = np.reshape(x, self.dims_fft)
        if self.fftshift_after.any():
            x = np.fft.ifftshift(x, axes=self.dirs[self.fftshift_after])
        if self.real:
            # Apply scaling to obtain a correct adjoint for this operator
            x = x.copy()
            x = np.swapaxes(x, -1, self.dirs[-1])
            x[..., 1 : 1 + (self.nffts[-1] - 1) // 2] /= np.sqrt(2)
            x = np.swapaxes(x, self.dirs[-1], -1)
            y = np.fft.irfft2(x, s=self.nffts, axes=self.dirs, **self._norm_kwargs)
        else:
            y = np.fft.ifft2(x, s=self.nffts, axes=self.dirs, **self._norm_kwargs)
        if self.norm is _FFTNorms.NONE:
            y *= self._scale
        if self.nffts[0] > self.dims[self.dirs[0]]:
            y = np.take(y, range(self.dims[self.dirs[0]]), axis=self.dirs[0])
        if self.nffts[1] > self.dims[self.dirs[1]]:
            y = np.take(y, range(self.dims[self.dirs[1]]), axis=self.dirs[1])
        if self.doifftpad:
            y = np.pad(y, self.ifftpad)
        if not self.clinear:
            y = np.real(y)
        y = y.astype(self.rdtype)
        if self.ifftshift_before.any():
            y = np.fft.fftshift(y, axes=self.dirs[self.ifftshift_before])
        return y.ravel()

    def __truediv__(self, y):
        if self.norm is not _FFTNorms.ORTHO:
            return self._rmatvec(y) / self._scale
        return self._rmatvec(y)


class _FFT2D_scipy(_BaseFFTND):
    """Two dimensional Fast-Fourier Transform using SciPy"""

    def __init__(
        self,
        dims,
        dirs=(0, 1),
        nffts=None,
        sampling=1.0,
        norm="ortho",
        real=False,
        ifftshift_before=False,
        fftshift_after=False,
        dtype="complex128",
    ):
        super().__init__(
            dims=dims,
            dirs=dirs,
            nffts=nffts,
            sampling=sampling,
            norm=norm,
            real=real,
            ifftshift_before=ifftshift_before,
            fftshift_after=fftshift_after,
            dtype=dtype,
        )

        # checks
        if self.ndim < 2:
            raise ValueError("FFT2D requires at least two input dimensions")
        if self.ndirs != 2:
            raise ValueError("FFT2D must be applied along exactly two dimensions")

        self.f1, self.f2 = self.fs
        del self.fs

        self._norm_kwargs = {"norm": None}  # equivalent to "backward" in Numpy/Scipy
        if self.norm is _FFTNorms.ORTHO:
            self._norm_kwargs["norm"] = "ortho"
        elif self.norm is _FFTNorms.NONE:
            self._scale = np.sqrt(np.prod(self.nffts))
        elif self.norm is _FFTNorms.ONE_OVER_N:
            self._scale = np.sqrt(1.0 / np.prod(self.nffts))

    def _matvec(self, x):
        x = np.reshape(x, self.dims)
        if self.ifftshift_before.any():
            x = scipy.fft.ifftshift(x, axes=self.dirs[self.ifftshift_before])
        if not self.clinear:
            x = np.real(x)
        if self.real:
            y = scipy.fft.rfft2(x, s=self.nffts, axes=self.dirs, **self._norm_kwargs)
            # Apply scaling to obtain a correct adjoint for this operator
            y = np.swapaxes(y, -1, self.dirs[-1])
            y[..., 1 : 1 + (self.nffts[-1] - 1) // 2] *= np.sqrt(2)
            y = np.swapaxes(y, self.dirs[-1], -1)
        else:
            y = scipy.fft.fft2(x, s=self.nffts, axes=self.dirs, **self._norm_kwargs)
        if self.norm is _FFTNorms.ONE_OVER_N:
            y *= self._scale
        if self.fftshift_after.any():
            y = scipy.fft.fftshift(y, axes=self.dirs[self.fftshift_after])
        return y.ravel()

    def _rmatvec(self, x):
        x = np.reshape(x, self.dims_fft)
        if self.fftshift_after.any():
            x = scipy.fft.ifftshift(x, axes=self.dirs[self.fftshift_after])
        if self.real:
            # Apply scaling to obtain a correct adjoint for this operator
            x = x.copy()
            x = np.swapaxes(x, -1, self.dirs[-1])
            x[..., 1 : 1 + (self.nffts[-1] - 1) // 2] /= np.sqrt(2)
            x = np.swapaxes(x, self.dirs[-1], -1)
            y = scipy.fft.irfft2(x, s=self.nffts, axes=self.dirs, **self._norm_kwargs)
        else:
            y = scipy.fft.ifft2(x, s=self.nffts, axes=self.dirs, **self._norm_kwargs)
        if self.norm is _FFTNorms.NONE:
            y *= self._scale
        y = np.take(y, range(self.dims[self.dirs[0]]), axis=self.dirs[0])
        y = np.take(y, range(self.dims[self.dirs[1]]), axis=self.dirs[1])
        if not self.clinear:
            y = np.real(y)
        if self.ifftshift_before.any():
            y = scipy.fft.fftshift(y, axes=self.dirs[self.ifftshift_before])
        return y.ravel()

    def __truediv__(self, y):
        if self.norm is not _FFTNorms.ORTHO:
            return self._rmatvec(y) / self._scale / self._scale
        return self._rmatvec(y)


def FFT2D(
    dims,
    dirs=(0, 1),
    nffts=None,
    sampling=1.0,
    norm="ortho",
    real=False,
    ifftshift_before=False,
    fftshift_after=False,
    dtype="complex128",
    engine="numpy",
):
    r"""Two dimensional Fast-Fourier Transform.

    Apply two dimensional Fast-Fourier Transform (FFT) to any pair of axes of a
    multi-dimensional array depending on the choice of ``dirs``.

    Using the default NumPy engine, the FFT operator is an overload to either the NumPy
    :py:func:`numpy.fft.fft2` (or :py:func:`numpy.fft.rfft2` for real models) in
    forward mode, and to :py:func:`numpy.fft.ifft2` (or :py:func:`numpy.fft.irfft2`
    for real models) in adjoint mode, or their CuPy equivalents.
    Alternatively, when the SciPy engine is chosen, the overloads are of
    :py:func:`scipy.fft.fft2` (or :py:func:`scipy.fft.rfft2` for real models) in
    forward mode, and to :py:func:`scipy.fft.ifft2` (or :py:func:`scipy.fft.irfft2`
    for real models) in adjoint mode.

    When using ``real=True``, the result of the forward is also multiplied by
    :math:`\sqrt{2}` for all frequency bins except zero and Nyquist, and the input of
    the adjoint is multiplied by :math:`1 / \sqrt{2}` for the same frequencies.

    For a real valued input signal, it is advised to use the flag ``real=True``
    as it stores the values of the Fourier transform of the last direction at positive
    frequencies only as values at negative frequencies are simply their complex conjugates.

    Parameters
    ----------
    dims : :obj:`tuple`
        Number of samples for each dimension
    dirs : :obj:`tuple`, optional
        Pair of directions along which FFT2D is applied
    nffts : :obj:`tuple` or :obj:`int`, optional
        Number of samples in Fourier Transform for each direction. In case only one
        dimension needs to be specified, use ``None`` for the other dimension in the
        tuple. The direction with None will use ``dims[dir]`` as ``nfft``. When
        supplying a tuple, the order must agree with that of ``dirs``. When a single
        value is passed, it will be used for both directions. As such the default is
        equivalent to ``nffts=(None, None)``.
    sampling : :obj:`tuple` or :obj:`float`, optional
        Sampling steps for each direction. When supplied a single value, it is used
        for both directions. Unlike ``nffts``, any ``None`` will not be converted to the
        default value.
    norm : `{"ortho", "none", "1/n"}`, optional
        .. versionadded:: 1.17.0

        - "ortho": Scales forward and adjoint FFT transforms with :math:`1/\sqrt{N_F}`,
          where :math:`N_F` is the number of samples in the Fourier domain given by
          product of all elements of ``nffts``.

        - "none": Does not scale the forward or the adjoint FFT transforms.

        - "1/n": Scales both the forward and adjoint FFT transforms by
          :math:`1/N_F`.

        .. note:: For "none" and "1/n", the operator is not unitary, that is, the
          adjoint is not the inverse. To invert the operator, simply use ``Op \ y``.

    real : :obj:`bool`, optional
        Model to which fft is applied has real numbers (``True``) or not
        (``False``). Used to enforce that the output of adjoint of a real
        model is real. Note that the real FFT is applied only to the first
        dimension to which the FFT2D operator is applied (last element of
        ``dirs``)
    ifftshift_before : :obj:`tuple` or :obj:`bool`, optional
        Apply ifftshift (``True``) or not (``False``) to model vector (before FFT).
        Consider using this option when the model vector's respective axis is symmetric
        with respect to the zero value sample. This will shift the zero value sample to
        coincide with the zero index sample. With such an arrangement, FFT will not
        introduce a sample-dependent phase-shift when compared to the continuous Fourier
        Transform.
        When passing a single value, the shift will the same for every direction. Pass
        a tuple to specify which dimensions are shifted.
    fftshift_after : :obj:`tuple` or :obj:`bool`, optional
        Apply fftshift (``True``) or not (``False``) to data vector (after FFT).
        Consider using this option when you require frequencies to be arranged
        naturally, from negative to positive. When not applying fftshift after FFT,
        frequencies are arranged from zero to largest positive, and then from negative
        Nyquist to the frequency bin before zero.
        When passing a single value, the shift will the same for every direction. Pass
        a tuple to specify which dimensions are shifted.
    engine : :obj:`str`, optional
        .. versionadded:: 1.17.0

        Engine used for fft computation (``numpy`` or ``scipy``).
    dtype : :obj:`str`, optional
        Type of elements in input array. Note that the ``dtype`` of the operator
        is the corresponding complex type even when a real type is provided.
        In addition, note that the NumPy backend does not support returning ``dtype``
        different than ``complex128``. As such, when using the NumPy backend, arrays will
        be force-casted to types corresponding to the supplied ``dtype``.
        The SciPy backend supports all precisions natively.
        Under both backends, when a real ``dtype`` is supplied, a real result will be
        enforced on the result of the ``rmatvec`` and the input of the ``matvec``.

    Attributes
    ----------
    dims_fft : :obj:`tuple`
        Shape of the array after the forward, but before linearization.

        For example, ``y_reshaped = (Op * x.ravel()).reshape(Op.dims_fft)``.
    f1 : :obj:`numpy.ndarray`
        Discrete Fourier Transform sample frequencies along ``dir[0]``
    f2 : :obj:`numpy.ndarray`
        Discrete Fourier Transform sample frequencies along ``dir[1]``
    real : :obj:`bool`
        When ``True``, uses ``rfft2``/``irfft2``
    rdtype : :obj:`bool`
        Expected input type to the forward
    cdtype : :obj:`bool`
        Output type of the forward. Complex equivalent to ``rdtype``.
    shape : :obj:`tuple`
        Operator shape
    clinear : :obj:`bool`
        .. versionadded:: 1.17.0

        Operator is complex-linear. Is false when either ``real=True`` or when
        ``dtype`` is not a complex type.
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly
        (``True``) or not (``False``)

    Raises
    ------
    ValueError
        - If ``dims`` has less than two elements.
        - If ``dirs`` does not have exactly two elements.
        - If ``nffts`` or ``sampling`` are not either a single value or a tuple with
          two elements.
        - If ``norm`` is not one of "ortho", "none", or "1/n".
    NotImplementedError
        If ``engine`` is neither ``numpy``, nor ``scipy``.

    See Also
    --------
    FFT: One-dimensional FFT
    FFTND: N-dimensional FFT

    Notes
    -----
    The FFT2D operator (using ``norm="ortho"``) applies the two-dimensional forward
    Fourier transform to a signal :math:`d(y, x)` in forward mode:

    .. math::
        D(k_y, k_x) = \mathscr{F} (d) = \frac{1}{\sqrt{N_F}} \iint\limits_{-\infty}^\infty d(y, x) e^{-j2\pi k_yy}
        e^{-j2\pi k_xx} \,\mathrm{d}y \,\mathrm{d}x

    Similarly, the  two-dimensional inverse Fourier transform is applied to
    the Fourier spectrum :math:`D(k_y, k_x)` in adjoint mode:

    .. math::
        d(y,x) = \mathscr{F}^{-1} (D) = \frac{1}{\sqrt{N_F}} \iint\limits_{-\infty}^\infty D(k_y, k_x) e^{j2\pi k_yy}
        e^{j2\pi k_xx} \,\mathrm{d}k_y  \,\mathrm{d}k_x

    where :math:`N_F` is the number of samples in the Fourier domain given by the
    product of the element of ``nffts``.
    Both operators are effectively discretized and solved by a fast iterative
    algorithm known as Fast Fourier Transform. Note that the FFT2D operator
    (using ``norm="ortho"``) is a special operator in that the adjoint is also
    the inverse of the forward mode. For other norms, this does not hold (see ``norm``
    help). However, for any norm, the 2D Fourier transform is Hermitian for real input
    signals.

    """
    if engine == "numpy":
        f = _FFT2D_numpy(
            dims=dims,
            dirs=dirs,
            nffts=nffts,
            sampling=sampling,
            norm=norm,
            real=real,
            ifftshift_before=ifftshift_before,
            fftshift_after=fftshift_after,
            dtype=dtype,
        )
    elif engine == "scipy":
        f = _FFT2D_scipy(
            dims=dims,
            dirs=dirs,
            nffts=nffts,
            sampling=sampling,
            norm=norm,
            real=real,
            ifftshift_before=ifftshift_before,
            fftshift_after=fftshift_after,
            dtype=dtype,
        )
    else:
        raise NotImplementedError("engine must be numpy or scipy")
    return f



import numpy as np

from pylops import LinearOperator


class Pad(LinearOperator):
    r"""Pad operator.

    Zero-pad model in forward model and extract non-zero subsequence
    in adjoint. Padding can be performed in one or multiple directions to any
    multi-dimensional input arrays.

    Parameters
    ----------
    dims : :obj:`int` or :obj:`tuple`
        Number of samples for each dimension
    pad : :obj:`tuple`
        Number of samples to pad. If ``dims`` is a scalar, ``pad`` is a single
        tuple ``(pad_in, pad_end)``. If ``dims`` is a tuple,
        ``pad`` is a tuple of tuples where each inner tuple contains
        the number of samples to pad in each dimension
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

    Raises
    ------
    ValueError
        If any element of ``pad`` is negative.

    Notes
    -----
    Given an array of size :math:`N`, the *Pad* operator simply adds
    :math:`\text{pad}_\text{in}` at the start and :math:`\text{pad}_\text{end}` at the end in forward mode:

    .. math::

        y_{i} = x_{i-\text{pad}_\text{in}}  \quad \forall
        i=\text{pad}_\text{in},\ldots,\text{pad}_\text{in}+N-1

    and :math:`y_i = 0 \quad \forall
    i=0,\ldots,\text{pad}_\text{in}-1, \text{pad}_\text{in}+N-1,\ldots,N+\text{pad}_\text{in}+\text{pad}_\text{end}`

    In adjoint mode, values from :math:`\text{pad}_\text{in}` to :math:`N-\text{pad}_\text{end}` are
    extracted from the data:

    .. math::

        x_{i} = y_{\text{pad}_\text{in}+i}  \quad \forall i=0, N-1

    """

    def __init__(self, dims, pad, dtype="float64"):
        if np.any(np.array(pad) < 0):
            raise ValueError("Padding must be positive or zero")
        self.dims = dims
        self.pad = pad
        self.reshape = False if isinstance(self.dims, int) else True
        if self.reshape:
            self.dimsd = [dim + p[0] + p[1] for dim, p in zip(dims, pad)]
        else:
            self.dimsd = dims + pad[0] + pad[1]
        self.shape = (np.prod(np.array(self.dimsd)), np.prod(np.array(self.dims)))
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        if self.reshape:
            y = x.reshape(self.dims)
            y = np.pad(y, self.pad, mode="constant")
        else:
            y = np.pad(x, self.pad, mode="constant")
        return y.ravel()

    def _rmatvec(self, x):
        if self.reshape:
            y = x.reshape(self.dimsd)
            for ax, pad in enumerate(self.pad):
                y = np.take(y, np.arange(pad[0], pad[0] + self.dims[ax]), axis=ax)
        else:
            y = x[self.pad[0] : self.pad[0] + self.dims]
        return y.ravel()



import os
from importlib import util

numba_enabled = util.find_spec("numba") is not None
cupy_enabled = (
    util.find_spec("cupy") is not None and int(os.getenv("CUPY_PYLOPS", 1)) == 1
)
cusignal_enabled = (
    util.find_spec("cusignal") is not None and int(os.getenv("CUSIGNAL_PYLOPS", 1)) == 1
)



import numpy as np

from pylops.utils.backend import get_module, to_numpy


def dottest(
    Op,
    nr=None,
    nc=None,
    tol=1e-6,
    complexflag=0,
    raiseerror=True,
    verb=False,
    backend="numpy",
):
    r"""Dot test.

    Generate random vectors :math:`\mathbf{u}` and :math:`\mathbf{v}`
    and perform dot-test to verify the validity of forward and adjoint
    operators. This test can help to detect errors in the operator
    implementation.

    Parameters
    ----------
    Op : :obj:`pylops.LinearOperator`
        Linear operator to test.
    nr : :obj:`int`
        Number of rows of operator (i.e., elements in data)
    nc : :obj:`int`
        Number of columns of operator (i.e., elements in model)
    tol : :obj:`float`, optional
        Dottest tolerance
    complexflag : :obj:`bool`, optional
        Generate random vectors with

        * ``0``: Real entries for model and data

        * ``1``: Complex entries for model and real entries for data

        * ``2``: Real entries for model and complex entries for data

        * ``3``: Complex entries for model and  data
    raiseerror : :obj:`bool`, optional
        Raise error or simply return ``False`` when dottest fails
    verb : :obj:`bool`, optional
        Verbosity
    backend : :obj:`str`, optional
        Backend used for dot test computations (``numpy`` or ``cupy``). This
        parameter will be used to choose how to create the random vectors.

    Raises
    ------
    ValueError
        If dot-test is not verified within chosen tolerance.

    Notes
    -----
    A dot-test is mathematical tool used in the development of numerical
    linear operators.

    More specifically, a correct implementation of forward and adjoint for
    a linear operator should verify the following *equality*
    within a numerical tolerance:

    .. math::
        (\mathbf{Op}\,\mathbf{u})^H\mathbf{v} =
        \mathbf{u}^H(\mathbf{Op}^H\mathbf{v})

    """
    ncp = get_module(backend)

    if nr is None:
        nr = Op.shape[0]
    if nc is None:
        nc = Op.shape[1]

    assert (nr, nc) == Op.shape, "Provided nr and nc do not match operator shape"

    # make u and v vectors
    if complexflag != 0:
        rdtype = np.real(np.ones(1, Op.dtype)).dtype

    if complexflag in (0, 2):
        u = ncp.random.randn(nc).astype(Op.dtype)
    else:
        u = ncp.random.randn(nc).astype(rdtype) + 1j * ncp.random.randn(nc).astype(
            rdtype
        )

    if complexflag in (0, 1):
        v = ncp.random.randn(nr).astype(Op.dtype)
    else:
        v = ncp.random.randn(nr).astype(rdtype) + 1j * ncp.random.randn(nr).astype(
            rdtype
        )

    y = Op.matvec(u)  # Op * u
    x = Op.rmatvec(v)  # Op'* v

    if getattr(Op, "clinear", True):
        yy = ncp.vdot(y, v)  # (Op  * u)' * v
        xx = ncp.vdot(u, x)  # u' * (Op' * v)
    else:
        # Op is only R-linear, so treat complex numbers as elements of R^2
        yy = ncp.dot(y.real, v.real) + ncp.dot(y.imag, v.imag)
        xx = ncp.dot(u.real, x.real) + ncp.dot(u.imag, x.imag)

    # convert back to numpy (in case cupy arrays were used), make into a numpy
    # array and extract the first element. This is ugly but allows to handle
    # complex numbers in subsequent prints also when using cupy arrays.
    xx, yy = np.array([to_numpy(xx)])[0], np.array([to_numpy(yy)])[0]

    # evaluate if dot test is passed
    if complexflag == 0:
        if np.abs((yy - xx) / ((yy + xx + 1e-15) / 2)) < tol:
            if verb:
                print("Dot test passed, v^T(Opu)=%f - u^T(Op^Tv)=%f" % (yy, xx))
            return True
        else:
            if raiseerror:
                raise ValueError(
                    "Dot test failed, v^T(Opu)=%f - u^T(Op^Tv)=%f" % (yy, xx)
                )
            if verb:
                print("Dot test failed, v^T(Opu)=%f - u^T(Op^Tv)=%f" % (yy, xx))
            return False
    else:
        # Check both real and imag parts
        checkreal = (
            np.abs(
                (np.real(yy) - np.real(xx)) / ((np.real(yy) + np.real(xx) + 1e-15) / 2)
            )
            < tol
        )
        checkimag = (
            np.abs(
                (np.imag(yy) - np.imag(xx)) / ((np.imag(yy) + np.imag(xx) + 1e-15) / 2)
            )
            < tol
        )
        if checkreal and checkimag:
            if verb:
                print(
                    "Dot test passed, v^T(Opu)=%f%+fi - u^T(Op^Tv)=%f%+fi"
                    % (yy.real, yy.imag, xx.real, xx.imag)
                )
            return True
        else:
            if raiseerror:
                raise ValueError(
                    "Dot test failed, v^H(Opu)=%f%+fi "
                    "- u^H(Op^Hv)=%f%+fi" % (yy.real, yy.imag, xx.real, xx.imag)
                )
            if verb:
                print(
                    "Dot test failed, v^H(Opu)=%f%+fi - u^H(Op^Hv)=%f%+fi"
                    % (yy.real, yy.imag, xx.real, xx.imag)
                )
            return False



import numpy as np
from scipy.ndimage import gaussian_filter

from pylops.utils.backend import get_array_module, get_toeplitz


def convmtx(h, n):
    r"""Convolution matrix

    Equivalent of `MATLAB's convmtx function
    <http://www.mathworks.com/help/signal/ref/convmtx.html>`_ .
    Makes a dense convolution matrix :math:`\mathbf{C}`
    such that the dot product ``np.dot(C, x)`` is the convolution of
    the filter :math:`h` and the input signal :math:`x`.

    Parameters
    ----------
    h : :obj:`np.ndarray`
        Convolution filter (1D array)
    n : :obj:`int`
        Number of columns (if :math:`\text{len}(h) < n`) or rows
        (if :math:`\text{len}(h) \geq n`) of convolution matrix

    Returns
    -------
    C : :obj:`np.ndarray`
        Convolution matrix of size :math:`\text{len}(h)+n-1 \times n`
        (if :math:`\text{len}(h) < n`) or :math:`n \times \text{len}(h)+n-1`
        (if :math:`\text{len}(h) \geq n`)

    """
    ncp = get_array_module(h)
    if len(h) < n:
        col_1 = ncp.r_[h[0], ncp.zeros(n - 1, dtype=h.dtype)]
        row_1 = ncp.r_[h, ncp.zeros(n - 1, dtype=h.dtype)]
    else:
        row_1 = ncp.r_[h[0], ncp.zeros(n - 1, dtype=h.dtype)]
        col_1 = ncp.r_[h, ncp.zeros(n - 1, dtype=h.dtype)]
    C = get_toeplitz(h)(col_1, row_1)
    return C


def nonstationary_convmtx(H, n, hc=0, pad=(0, 0)):
    r"""Convolution matrix from a bank of filters

    Makes a dense convolution matrix :math:`\mathbf{C}`
    such that the dot product ``np.dot(C, x)`` is the nonstationary
    convolution of the bank of filters :math:`H=[h_1, h_2, h_n]`
    and the input signal :math:`x`.

    Parameters
    ----------
    H : :obj:`np.ndarray`
        Convolution filters (2D array of shape
        :math:`[n_\text{filters} \times n_{h}]`
    n : :obj:`int`
        Number of columns of convolution matrix
    hc : :obj:`np.ndarray`, optional
        Index of center of first filter
    pad : :obj:`np.ndarray`
        Zero-padding to apply to the bank of filters before and after the
        provided values (use it to avoid wrap-around or pass filters with
        enough padding)

    Returns
    -------
    C : :obj:`np.ndarray`
        Convolution matrix

    """
    ncp = get_array_module(H)

    H = ncp.pad(H, ((0, 0), pad), mode="constant")
    C = ncp.array([ncp.roll(h, ih) for ih, h in enumerate(H)])
    C = C[:, pad[0] + hc : pad[0] + hc + n].T  # take away edges
    return C


def slope_estimate(d, dz=1.0, dx=1.0, smooth=5, eps=0):
    r"""Local slope estimation

    Local slopes are estimated using the *Structure Tensor* algorithm [1]_.
    Slopes are returned as :math:`\tan\theta`, defined
    in a RHS coordinate system with :math:`z`-axis pointing upward.

    .. note:: For stability purposes, it is important to ensure that the orders
        of magnitude of the samplings are similar.

    Parameters
    ----------
    d : :obj:`np.ndarray`
        Input dataset of size :math:`n_z \times n_x`
    dz : :obj:`float`
        Sampling in :math:`z`-axis, :math:`\Delta z`

        .. warning::
            Since version 1.17.0, defaults to 1.0.

    dx : :obj:`float`
        Sampling in :math:`x`-axis, :math:`\Delta x`

        .. warning::
            Since version 1.17.0, defaults to 1.0.

    smooth : :obj:`float`, optional
        Standard deviation for Gaussian kernel. The standard deviations of the
        Gaussian filter are given for each axis as a sequence, or as a single number,
        in which case it is equal for all axes.

        .. warning::
            Default changed in version 1.17.0 to 5 from previous value of 20.

    eps : :obj:`float`, optional
        .. versionadded:: 1.17.0

        Regularization term. All slopes where :math:`|g_{zx}| < \epsilon \max |g_{zx}|`
        are set to zero. All anisotropies where :math:`\lambda_\text{max} < \epsilon`
        are also set to zero. See Notes. When using with small values of ``smooth``,
        start from a very small number (e.g. 1e-10) and start increasing by a power
        of 10 until results are satisfactory.

    Returns
    -------
    slopes : :obj:`np.ndarray`
        Estimated local slopes. Unit is that of :math:`\Delta z/\Delta x`.

        .. warning::
            Prior to version 1.17.0, erroneously returned angles in radians instead of
            slopes.

    anisotropies : :obj:`np.ndarray`
        Estimated local anisotropies: :math:`1-\lambda_\text{min}/\lambda_\text{max}`

        .. note::
            Since 1.17.0, changed name from ``linearity`` to ``anisotropies``.
            Definition remains the same.


    Notes
    -----
    For each pixel of the input dataset :math:`\mathbf{d}` the local gradients
    :math:`g_z = \frac{\partial \mathbf{d}}{\partial z}` and
    :math:`g_x = \frac{\partial \mathbf{d}}{\partial x}` are computed
    and used to define the following three quantities:

    .. math::
        \begin{align}
        g_{zz} &= \left(\frac{\partial \mathbf{d}}{\partial z}\right)^2\\
        g_{xx} &= \left(\frac{\partial \mathbf{d}}{\partial x}\right)^2\\
        g_{zx} &= \frac{\partial \mathbf{d}}{\partial z}\cdot\frac{\partial \mathbf{d}}{\partial x}
        \end{align}

    They are then spatially smoothed and at each pixel their smoothed versions are
    arranged in a :math:`2 \times 2` matrix called the *smoothed
    gradient-square tensor*:

    .. math::
        \mathbf{G} =
        \begin{bmatrix}
           g_{zz}  & g_{zx} \\
           g_{zx}  & g_{xx}
        \end{bmatrix}

    Local slopes can be expressed as
    :math:`p = \frac{\lambda_\text{max} - g_{zz}}{g_{zx}}`,
    where :math:`\lambda_\text{max}` is the largest eigenvalue of :math:`\mathbf{G}`.

    Moreover, we can obtain a measure of local anisotropy, defined as

    .. math::
        a = 1-\lambda_\text{min}/\lambda_\text{max}

    where :math:`\lambda_\text{min}` is the smallest eigenvalue of :math:`\mathbf{G}`.
    A value of :math:`a = 0`  indicates perfect isotropy whereas :math:`a = 1`
    indicates perfect anisotropy.

    .. [1] Van Vliet, L. J.,  Verbeek, P. W., "Estimators for orientation and
        anisotropy in digitized images", Journal ASCI Imaging Workshop. 1995.

    """
    slopes = np.zeros_like(d)
    anisos = np.zeros_like(d)

    gz, gx = np.gradient(d, dz, dx)
    gzz, gzx, gxx = gz * gz, gz * gx, gx * gx

    # smoothing
    gzz = gaussian_filter(gzz, sigma=smooth)
    gzx = gaussian_filter(gzx, sigma=smooth)
    gxx = gaussian_filter(gxx, sigma=smooth)

    gmax = np.max(np.abs(gzx))
    if gmax == 0.0:
        return slopes, anisos

    gzz /= gmax
    gzx /= gmax
    gxx /= gmax

    lcommon1 = 0.5 * (gzz + gxx)
    lcommon2 = 0.5 * np.sqrt((gzz - gxx) ** 2 + 4 * gzx ** 2)
    l1 = lcommon1 + lcommon2
    l2 = lcommon1 - lcommon2

    regdata = np.abs(gzx) > eps
    slopes[regdata] = (l1 - gzz)[regdata] / gzx[regdata]

    regdata = np.abs(l1) > eps
    anisos[regdata] = 1 - l2[regdata] / l1[regdata]

    return slopes, anisos



from __future__ import division

import logging

import numpy as np
import scipy as sp
from scipy.linalg import eigvals, lstsq, solve
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import LinearOperator as spLinearOperator
from scipy.sparse.linalg import eigs as sp_eigs
from scipy.sparse.linalg import eigsh as sp_eigsh
from scipy.sparse.linalg import lobpcg as sp_lobpcg
from scipy.sparse.linalg import lsqr, spsolve
from scipy.sparse.linalg.interface import _ProductLinearOperator

from pylops.optimization.solver import cgls
from pylops.utils.backend import get_array_module, get_module, get_sparse_eye
from pylops.utils.estimators import trace_hutchinson, trace_hutchpp, trace_nahutchpp

logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.WARNING)


class LinearOperator(spLinearOperator):
    """Common interface for performing matrix-vector products.

    This class is an overload of the
    :py:class:`scipy.sparse.linalg.LinearOperator` class. It adds
    functionalities by overloading standard operators such as ``__truediv__``
    as well as creating convenience methods such as ``eigs``, ``cond``, and
    ``conj``.

    .. note:: End users of PyLops should not use this class directly but simply
      use operators that are already implemented. This class is meant for
      developers and it has to be used as the parent class of any new operator
      developed within PyLops. Find more details regarding implementation of
      new operators at :ref:`addingoperator`.

    Parameters
    ----------
    Op : :obj:`scipy.sparse.linalg.LinearOperator` or :obj:`scipy.sparse.linalg._ProductLinearOperator` or :obj:`scipy.sparse.linalg._SumLinearOperator`
        Operator
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly
        (``True``) or not (``False``)
    clinear : :obj:`bool`
        .. versionadded:: 1.17.0

        Operator is complex-linear.
    """

    def __init__(self, Op=None, explicit=False, clinear=None):
        self.explicit = explicit
        if Op is not None:
            self.Op = Op
            self.shape = self.Op.shape
            self.dtype = self.Op.dtype
            self.clinear = getattr(self.Op, "clinear", True)
        if clinear is not None:
            self.clinear = clinear

    def _matvec(self, x):
        if callable(self.Op._matvec):
            return self.Op._matvec(x)

    def _rmatvec(self, x):
        if callable(self.Op._rmatvec):
            return self.Op._rmatvec(x)

    def _matmat(self, X):
        """Matrix-matrix multiplication handler.

        Modified version of scipy _matmat to avoid having trailing dimension
        in col when provided to matvec
        """
        if sp.sparse.issparse(X):
            y = np.vstack([self.matvec(col.toarray().reshape(-1)) for col in X.T]).T
        else:
            y = np.vstack([self.matvec(col.reshape(-1)) for col in X.T]).T
        return y

    def __mul__(self, x):
        y = super().__mul__(x)
        if isinstance(y, spLinearOperator):
            y = aslinearoperator(y)
        return y

    def __rmul__(self, x):
        if np.isscalar(x):
            return aslinearoperator(_ScaledLinearOperator(self, x))
        else:
            return NotImplemented

    def __pow__(self, p):
        return aslinearoperator(super().__pow__(p))

    def __add__(self, x):
        return aslinearoperator(super().__add__(x))

    def __neg__(self):
        return aslinearoperator(_ScaledLinearOperator(self, -1))

    def __sub__(self, x):
        return aslinearoperator(super().__sub__(x))

    def _adjoint(self):
        return aslinearoperator(super()._adjoint())

    def _transpose(self):
        return aslinearoperator(super()._transpose())

    def matvec(self, x):
        """Matrix-vector multiplication.

        Modified version of scipy matvec which does not consider the case
        where the input vector is ``np.matrix`` (the use ``np.matrix`` is now
        discouraged in numpy's documentation).

        Parameters
        ----------
        x : :obj:`numpy.ndarray`
            Input array of shape (N,) or (N,1)

        Returns
        -------
        y : :obj:`numpy.ndarray`
            Output array of shape (M,) or (M,1)

        """
        M, N = self.shape

        if x.shape != (N,) and x.shape != (N, 1):
            raise ValueError("dimension mismatch")

        y = self._matvec(x)

        if x.ndim == 1:
            y = y.reshape(M)
        elif x.ndim == 2:
            y = y.reshape(M, 1)
        else:
            raise ValueError("invalid shape returned by user-defined matvec()")
        return y

    def rmatvec(self, x):
        """Adjoint matrix-vector multiplication.

        Modified version of scipy rmatvec which does not consider the case
        where the input vector is ``np.matrix`` (the use ``np.matrix`` is now
        discouraged in numpy's documentation).

        Parameters
        ----------
        y : :obj:`numpy.ndarray`
            Input array of shape (M,) or (M,1)

        Returns
        -------
        x : :obj:`numpy.ndarray`
            Output array of shape (N,) or (N,1)

        """
        M, N = self.shape

        if x.shape != (M,) and x.shape != (M, 1):
            raise ValueError("dimension mismatch")

        y = self._rmatvec(x)

        if x.ndim == 1:
            y = y.reshape(N)
        elif x.ndim == 2:
            y = y.reshape(N, 1)
        else:
            raise ValueError("invalid shape returned by user-defined rmatvec()")
        return y

    def matmat(self, X):
        """Matrix-matrix multiplication.

        Modified version of scipy matmat which does not consider the case
        where the input vector is ``np.matrix`` (the use ``np.matrix`` is now
        discouraged in numpy's documentation).

        Parameters
        ----------
        x : :obj:`numpy.ndarray`
            Input array of shape (N,K)

        Returns
        -------
        y : :obj:`numpy.ndarray`
            Output array of shape (M,K)

        """
        if X.ndim != 2:
            raise ValueError("expected 2-d ndarray or matrix, " "not %d-d" % X.ndim)
        if X.shape[0] != self.shape[1]:
            raise ValueError("dimension mismatch: %r, %r" % (self.shape, X.shape))
        Y = self._matmat(X)
        return Y

    def rmatmat(self, X):
        """Matrix-matrix multiplication.

        Modified version of scipy rmatmat which does not consider the case
        where the input vector is ``np.matrix`` (the use ``np.matrix`` is now
        discouraged in numpy's documentation).

        Parameters
        ----------
        x : :obj:`numpy.ndarray`
            Input array of shape (M,K)

        Returns
        -------
        y : :obj:`numpy.ndarray`
            Output array of shape (N,K)

        """
        if X.ndim != 2:
            raise ValueError("expected 2-d ndarray or matrix, " "not %d-d" % X.ndim)
        if X.shape[0] != self.shape[0]:
            raise ValueError("dimension mismatch: %r, %r" % (self.shape, X.shape))
        Y = self._rmatmat(X)
        return Y

    def dot(self, x):
        """Matrix-matrix or matrix-vector multiplication.

        Parameters
        ----------
        x : np.ndarray
            Input array (or matrix)

        Returns
        -------
        y : np.ndarray
            Output array (or matrix) that represents
            the result of applying the linear operator on x.

        """
        if isinstance(x, LinearOperator):
            Op = _ProductLinearOperator(self, x)
            # Output is C-Linear only if both operators are
            Op.clinear = getattr(self, "clinear", True) and getattr(x, "clinear", True)
            return Op
        elif np.isscalar(x):
            return _ScaledLinearOperator(self, x)
        else:
            if x.ndim == 1:  # or x.ndim == 2 and x.shape[1] == 1:
                return self.matvec(x)
            elif x.ndim == 2:
                return self.matmat(x)
            else:
                raise ValueError("expected 1-d or 2-d array or matrix, got %r" % x)

    def div(self, y, niter=100):
        r"""Solve the linear problem :math:`\mathbf{y}=\mathbf{A}\mathbf{x}`.

        Overloading of operator ``/`` to improve expressivity of `Pylops`
        when solving inverse problems.

        Parameters
        ----------
        y : :obj:`np.ndarray`
            Data
        niter : :obj:`int`, optional
            Number of iterations (to be used only when ``explicit=False``)

        Returns
        -------
        xest : :obj:`np.ndarray`
            Estimated model

        """
        xest = self.__truediv__(y, niter=niter)
        return xest

    def __truediv__(self, y, niter=100):
        if self.explicit is True:
            if sp.sparse.issparse(self.A):
                # use scipy solver for sparse matrices
                xest = spsolve(self.A, y)
            elif isinstance(self.A, np.ndarray):
                # use scipy solvers for dense matrices (used for backward
                # compatibility, could be switched to numpy equivalents)
                if self.A.shape[0] == self.A.shape[1]:
                    xest = solve(self.A, y)
                else:
                    xest = lstsq(self.A, y)[0]
            else:
                # use numpy/cupy solvers for dense matrices
                ncp = get_array_module(y)
                if self.A.shape[0] == self.A.shape[1]:
                    xest = ncp.linalg.solve(self.A, y)
                else:
                    xest = ncp.linalg.lstsq(self.A, y)[0]
        else:
            if isinstance(y, np.ndarray):
                # numpy backend
                xest = lsqr(self, y, iter_lim=niter)[0]
            else:
                # cupy backend
                ncp = get_array_module(y)
                xest = cgls(
                    self,
                    y,
                    x0=ncp.zeros(int(self.shape[1]), dtype=self.dtype),
                    niter=niter,
                )[0]
        return xest

    def todense(self, backend="numpy"):
        r"""Return dense matrix.

        The operator is converted into its dense matrix equivalent. In order
        to do so, square or tall operators are applied to an identity matrix
        whose number of rows and columns is equivalent to the number of
        columns of the operator. Conversely, for skinny operators, the
        transpose operator is applied to an identity matrix
        whose number of rows and columns is equivalent to the number of
        rows of the operator and the resulting matrix is transposed
        (and complex conjugated).

        Note that this operation may be costly for operators with large number
        of rows and columns and it should be used mostly as a way to inspect
        the structure of the matricial equivalent of the operator.

        Parameters
        ----------
        backend : :obj:`str`, optional
            Backend used to densify matrix (``numpy`` or ``cupy``). Note that
            this must be consistent with how the operator has been created.

        Returns
        -------
        matrix : :obj:`numpy.ndarray` or :obj:`cupy.ndarray`
            Dense matrix.

        """
        ncp = get_module(backend)

        # Wrap self into a LinearOperator. This is done for cases where self
        # is a _SumLinearOperator or _ProductLinearOperator, so that it regains
        # the dense method
        Op = aslinearoperator(self)

        # Create identity matrix
        shapemin = min(Op.shape)
        if shapemin <= 1e3:
            # use numpy for small matrices (faster but heavier on memory)
            identity = ncp.eye(shapemin, dtype=self.dtype)
        else:
            # use scipy for small matrices (slower but lighter on memory)
            identity = get_sparse_eye(ncp.ones(1))(shapemin, dtype=self.dtype).tocsc()

        # Apply operator
        if Op.shape[1] == shapemin:
            matrix = Op.matmat(identity)
        else:
            matrix = np.conj(Op.rmatmat(identity)).T
        return matrix

    def tosparse(self):
        r"""Return sparse matrix.

        The operator in converted into its sparse (CSR) matrix equivalent. In order
        to do so, the operator is applied to series of unit vectors with length equal
        to the number of coloumns in the original operator.

        Returns
        -------
        matrix : :obj:`scipy.sparse.csr_matrix`
            Sparse matrix.

        """
        Op = aslinearoperator(self)
        (_, n) = self.shape

        # stores non-zero data for the sparse matrix creation
        entries = []
        indices = []

        # loop through columns of self
        for i in range(n):

            # make a unit vector for the ith column
            unit_i = np.zeros(n)
            unit_i[i] = 1

            # multiply unit vector to self and find the non-zeros
            res_i = Op * unit_i
            rows_nz = np.where(res_i != 0)[0]

            # append the non-zero values and indices to the lists
            for j in rows_nz:
                indices.append([i, j])
            entries_i = res_i[rows_nz]
            for e in entries_i:
                entries.append(e)

        # post process the entries / indices for scipy.sparse.csr_matrix
        entries = np.array(entries)
        indices = np.array(indices)
        i, j = indices[:, 0], indices[:, 1]

        # construct a sparse, CSR matrix from the entries / indices data.
        matrix = csr_matrix((entries, (j, i)), shape=self.shape, dtype=self.dtype)
        return matrix

    def eigs(
        self, neigs=None, symmetric=False, niter=None, uselobpcg=False, **kwargs_eig
    ):
        r"""Most significant eigenvalues of linear operator.

        Return an estimate of the most significant eigenvalues
        of the linear operator. If the operator has rectangular
        shape (``shape[0]!=shape[1]``), eigenvalues are first
        computed for the square operator :math:`\mathbf{A^H}\mathbf{A}`
        and the square-root values are returned.

        Parameters
        ----------
        neigs : :obj:`int`
            Number of eigenvalues to compute (if ``None``, return all). Note
            that for ``explicit=False``, only :math:`N-1` eigenvalues can be
            computed where :math:`N` is the size of the operator in the
            model space
        symmetric : :obj:`bool`, optional
            Operator is symmetric (``True``) or not (``False``). User should
            set this parameter to ``True`` only when it is guaranteed that the
            operator is real-symmetric or complex-hermitian matrices
        niter : :obj:`int`, optional
            Number of iterations for eigenvalue estimation
        uselobpcg : :obj:`bool`, optional
            Use :func:`scipy.sparse.linalg.lobpcg`
        **kwargs_eig
            Arbitrary keyword arguments for :func:`scipy.sparse.linalg.eigs`,
            :func:`scipy.sparse.linalg.eigsh`, or
            :func:`scipy.sparse.linalg.lobpcg`

        Returns
        -------
        eigenvalues : :obj:`numpy.ndarray`
            Operator eigenvalues.

        Raises
        -------
        ValueError
            If ``uselobpcg=True`` for a non-symmetric square matrix with
            complex type

        Notes
        -----
        Depending on the size of the operator, whether it is explicit or not
        and the number of eigenvalues requested, different algorithms are
        used by this routine.

        More precisely, when only a limited number of eigenvalues is requested
        the :func:`scipy.sparse.linalg.eigsh` method is used in case of
        ``symmetric=True`` and the :func:`scipy.sparse.linalg.eigs` method
        is used ``symmetric=False``. However, when the matrix is represented
        explicitly within the linear operator (``explicit=True``) and all the
        eigenvalues are requested the :func:`scipy.linalg.eigvals` is used
        instead.

        Finally, when only a limited number of eigenvalues is required,
        it is also possible to explicitly choose to use the
        ``scipy.sparse.linalg.lobpcg`` method via the ``uselobpcg`` input
        parameter flag.

        Most of these algorithms are a port of ARPACK [1]_, a Fortran package
        which provides routines for quickly finding eigenvalues/eigenvectors
        of a matrix. As ARPACK requires only left-multiplication
        by the matrix in question, eigenvalues/eigenvectors can also be
        estimated for linear operators when the dense matrix is not available.

        .. [1] http://www.caam.rice.edu/software/ARPACK/

        """
        if self.explicit and isinstance(self.A, np.ndarray):
            if self.shape[0] == self.shape[1]:
                if neigs is None or neigs == self.shape[1]:
                    eigenvalues = eigvals(self.A)
                else:
                    if not symmetric and np.iscomplexobj(self) and uselobpcg:
                        raise ValueError(
                            "cannot use scipy.sparse.linalg.lobpcg "
                            "for non-symmetric square matrices of "
                            "complex type..."
                        )
                    if symmetric and uselobpcg:
                        X = np.random.rand(self.shape[0], neigs).astype(self.dtype)
                        eigenvalues = sp_lobpcg(
                            self.A, X=X, maxiter=niter, **kwargs_eig
                        )[0]
                    elif symmetric:
                        eigenvalues = sp_eigsh(
                            self.A, k=neigs, maxiter=niter, **kwargs_eig
                        )[0]
                    else:
                        eigenvalues = sp_eigs(
                            self.A, k=neigs, maxiter=niter, **kwargs_eig
                        )[0]

            else:
                if neigs is None or neigs == self.shape[1]:
                    eigenvalues = np.sqrt(eigvals(np.dot(np.conj(self.A.T), self.A)))
                else:
                    if uselobpcg:
                        X = np.random.rand(self.shape[1], neigs).astype(self.dtype)
                        eigenvalues = np.sqrt(
                            sp_lobpcg(
                                np.dot(np.conj(self.A.T), self.A),
                                X=X,
                                maxiter=niter,
                                **kwargs_eig,
                            )[0]
                        )
                    else:
                        eigenvalues = np.sqrt(
                            sp_eigsh(
                                np.dot(np.conj(self.A.T), self.A),
                                k=neigs,
                                maxiter=niter,
                                **kwargs_eig,
                            )[0]
                        )
        else:
            if neigs is None or neigs >= self.shape[1]:
                neigs = self.shape[1] - 2
            if self.shape[0] == self.shape[1]:
                if not symmetric and np.iscomplexobj(self) and uselobpcg:
                    raise ValueError(
                        "cannot use scipy.sparse.linalg.lobpcg for "
                        "non symmetric square matrices of "
                        "complex type..."
                    )
                if symmetric and uselobpcg:
                    X = np.random.rand(self.shape[0], neigs).astype(self.dtype)
                    eigenvalues = sp_lobpcg(self, X=X, maxiter=niter, **kwargs_eig)[0]
                elif symmetric:
                    eigenvalues = sp_eigsh(self, k=neigs, maxiter=niter, **kwargs_eig)[
                        0
                    ]
                else:
                    eigenvalues = sp_eigs(self, k=neigs, maxiter=niter, **kwargs_eig)[0]
            else:
                if uselobpcg:
                    X = np.random.rand(self.shape[1], neigs).astype(self.dtype)
                    eigenvalues = np.sqrt(
                        sp_lobpcg(self.H * self, X=X, maxiter=niter, **kwargs_eig)[0]
                    )
                else:
                    eigenvalues = np.sqrt(
                        sp_eigs(self.H * self, k=neigs, maxiter=niter, **kwargs_eig)[0]
                    )
        return -np.sort(-eigenvalues)

    def cond(self, uselobpcg=False, **kwargs_eig):
        r"""Condition number of linear operator.

        Return an estimate of the condition number of the linear operator as
        the ratio of the largest and lowest estimated eigenvalues.

        Parameters
        ----------
        uselobpcg : :obj:`bool`, optional
            Use :func:`scipy.sparse.linalg.lobpcg` to compute eigenvalues
        **kwargs_eig
            Arbitrary keyword arguments for :func:`scipy.sparse.linalg.eigs`,
            :func:`scipy.sparse.linalg.eigsh`, or
            :func:`scipy.sparse.linalg.lobpcg`

        Returns
        -------
        eigenvalues : :obj:`numpy.ndarray`
            Operator eigenvalues.

        Notes
        -----
        The condition number of a matrix (or linear operator) can be estimated
        as the ratio of the largest and lowest estimated eigenvalues:

        .. math::
            k= \frac{\lambda_{max}}{\lambda_{min}}

        The condition number provides an indication of the rate at which the
        solution of the inversion of the linear operator :math:`A` will
        change with respect to a change in the data :math:`y`.

        Thus, if the condition number is large, even a small error in :math:`y`
        may cause a large error in :math:`x`. On the other hand, if the
        condition number is small then the error in :math:`x` is not much
        bigger than the error in :math:`y`. A problem with a low condition
        number is said to be *well-conditioned*, while a problem with a high
        condition number is said to be *ill-conditioned*.

        """
        if not uselobpcg:
            cond = (
                self.eigs(neigs=1, which="LM", **kwargs_eig).item()
                / self.eigs(neigs=1, which="SM", **kwargs_eig).item()
            )
        else:
            cond = (
                self.eigs(neigs=1, uselobpcg=True, largest=True, **kwargs_eig).item()
                / self.eigs(neigs=1, uselobpcg=True, largest=False, **kwargs_eig).item()
            )

        return cond

    def conj(self):
        """Complex conjugate operator

        Returns
        -------
        conjop : :obj:`pylops.LinearOperator`
            Complex conjugate operator

        """
        conjop = _ConjLinearOperator(self)
        return conjop

    def apply_columns(self, cols):
        """Apply subset of columns of operator

        This method can be used to wrap a LinearOperator and mimic the action
        of a subset of columns of the operator on a reduced model in forward
        mode, and retrieve only the result of a subset of rows in adjoint mode.

        Note that unless the operator has ``explicit=True``, this is not
        optimal as the entire forward and adjoint passes of the original
        operator will have to be perfomed. It can however be useful for the
        implementation of solvers such as Orthogonal Matching Pursuit (OMP)
        that iteratively build a solution by evaluate only a subset of the
        columns of the operator.

        Parameters
        ----------
        cols : :obj:`list`
            Columns to be selected

        Returns
        -------
        colop : :obj:`pylops.LinearOperator`
            Apply column operator

        """
        colop = _ColumnLinearOperator(self, cols)
        return colop

    def toreal(self, forw=True, adj=True):
        """Real operator

        Parameters
        ----------
        forw : :obj:`bool`, optional
            Apply real to output of forward pass
        adj : :obj:`bool`, optional
            Apply real to output of adjoint pass

        Returns
        -------
        realop : :obj:`pylops.LinearOperator`
            Real operator

        """
        realop = _RealImagLinearOperator(self, forw, adj, True)
        return realop

    def toimag(self, forw=True, adj=True):
        """Imag operator

        Parameters
        ----------
        forw : :obj:`bool`, optional
            Apply imag to output of forward pass
        adj : :obj:`bool`, optional
            Apply imag to output of adjoint pass

        Returns
        -------
        imagop : :obj:`pylops.LinearOperator`
            Imag operator

        """
        imagop = _RealImagLinearOperator(self, forw, adj, False)
        return imagop

    def trace(
        self,
        neval=None,
        method=None,
        backend="numpy",
        **kwargs_trace,
    ):
        r"""Trace of linear operator.

        Returns the trace (or its estimate) of the linear operator.

        Parameters
        ----------
        neval : :obj:`int`, optional
            Maximum number of matrix-vector products compute. Default depends
            ``method``.
        method : :obj:`str`, optional
            Should be one of the following:

                - **explicit**: If the operator is not explicit, will convert to
                  dense first.
                - **hutchinson**: see :obj:`pylops.utils.trace_hutchinson`
                - **hutch++**: see :obj:`pylops.utils.trace_hutchpp`
                - **na-hutch++**: see :obj:`pylops.utils.trace_nahutchpp`

            Defaults to 'explicit' for explicit operators, and 'Hutch++' for
            the rest.
        backend : :obj:`str`, optional
            Backend used to densify matrix (``numpy`` or ``cupy``). Note that
            this must be consistent with how the operator has been created.
        **kwargs_trace
            Arbitrary keyword arguments passed to
            :obj:`pylops.utils.trace_hutchinson`,
            :obj:`pylops.utils.trace_hutchpp`, or
            :obj:`pylops.utils.trace_nahutchpp`

        Returns
        -------
        trace : :obj:`self.dtype`
            Operator trace.

        Raises
        -------
        ValueError
             If the operator has rectangular shape (``shape[0] != shape[1]``)

        NotImplementedError
            If the ``method`` is not one of the available methods.
        """
        if self.shape[0] != self.shape[1]:
            raise ValueError("operator is not square.")

        ncp = get_module(backend)

        if method is None:
            method = "explicit" if self.explicit else "hutch++"

        method_l = method.lower()
        if method_l == "explicit":
            A = self.A if self.explicit else self.todense(backend=backend)
            return ncp.trace(A)
        elif method_l == "hutchinson":
            return trace_hutchinson(self, neval=neval, backend=backend, **kwargs_trace)
        elif method_l == "hutch++":
            return trace_hutchpp(self, neval=neval, backend=backend, **kwargs_trace)
        elif method_l == "na-hutch++":
            return trace_nahutchpp(self, neval=neval, backend=backend, **kwargs_trace)
        else:
            raise NotImplementedError(f"method {method} not available.")


def _get_dtype(operators, dtypes=None):
    if dtypes is None:
        dtypes = []
    opdtypes = []
    for obj in operators:
        if obj is not None and hasattr(obj, "dtype"):
            opdtypes.append(obj.dtype)
    return np.find_common_type(opdtypes, dtypes)


class _ScaledLinearOperator(spLinearOperator):
    """
    Sum Linear Operator

    Modified version of scipy _ScaledLinearOperator which uses a modified
    _get_dtype where the scalar and operator types are passed separately to
    np.find_common_type. Passing them together does lead to problems when using
    np.float32 operators which are cast to np.float64

    """

    def __init__(self, A, alpha):
        if not isinstance(A, spLinearOperator):
            raise ValueError("LinearOperator expected as A")
        if not np.isscalar(alpha):
            raise ValueError("scalar expected as alpha")
        dtype = _get_dtype([A], [type(alpha)])
        super(_ScaledLinearOperator, self).__init__(dtype, A.shape)
        self.args = (A, alpha)

    def _matvec(self, x):
        return self.args[1] * self.args[0].matvec(x)

    def _rmatvec(self, x):
        return np.conj(self.args[1]) * self.args[0].rmatvec(x)

    def _rmatmat(self, x):
        return np.conj(self.args[1]) * self.args[0].rmatmat(x)

    def _matmat(self, x):
        return self.args[1] * self.args[0].matmat(x)

    def _adjoint(self):
        A, alpha = self.args
        return A.H * np.conj(alpha)


class _ConjLinearOperator(LinearOperator):
    """Complex conjugate linear operator"""

    def __init__(self, Op):
        if not isinstance(Op, spLinearOperator):
            raise TypeError("Op must be a LinearOperator")
        super(_ConjLinearOperator, self).__init__(Op, Op.shape)
        self.Op = Op

    def _matvec(self, x):
        return (self.Op._matvec(x.conj())).conj()

    def _rmatvec(self, x):
        return (self.Op._rmatvec(x.conj())).conj()

    def _adjoint(self):
        return _ConjLinearOperator(self.Op.H)


class _ColumnLinearOperator(LinearOperator):
    """Column selector linear operator

    Produces the forward and adjoint passes with a subset of columns of the
    original operator
    """

    def __init__(self, Op, cols):
        if not isinstance(Op, spLinearOperator):
            raise TypeError("Op must be a LinearOperator")
        super(_ColumnLinearOperator, self).__init__(Op, Op.explicit)
        self.Op = Op
        self.cols = cols
        self.shape = (Op.shape[0], len(cols))
        if self.explicit:
            self.Opcol = Op.A[:, cols]

    def _matvec(self, x):
        ncp = get_array_module(x)
        if self.explicit:
            y = self.Opcol @ x
        else:
            y = ncp.zeros(int(self.Op.shape[1]), dtype=self.dtype)
            y[self.cols] = x
            y = self.Op._matvec(y)
        return y

    def _rmatvec(self, x):
        if self.explicit:
            y = self.Opcol.T.conj() @ x
        else:
            y = self.Op._rmatvec(x)
            y = y[self.cols]
        return y


class _RealImagLinearOperator(LinearOperator):
    """Real-Imag linear operator

    Computes forward and adjoint passes of an operator Op and returns only
    its real (or imaginary) component. Note that for the adjoint step the
    output must be complex conjugated (i.e. opposite of the imaginary part is
    returned)
    """

    def __init__(self, Op, forw=True, adj=True, real=True):
        if not isinstance(Op, spLinearOperator):
            raise TypeError("Op must be a LinearOperator")
        super(_RealImagLinearOperator, self).__init__(Op, Op.shape)
        self.Op = Op
        self.real = real
        self.forw = forw
        self.adj = adj
        self.dtype = np.array(0, dtype=self.Op.dtype).real.dtype

    def _matvec(self, x):
        ncp = get_array_module(x)
        y = self.Op._matvec(x)
        if self.forw:
            if self.real:
                y = ncp.real(y)
            else:
                y = ncp.imag(y)
        return y

    def _rmatvec(self, x):
        ncp = get_array_module(x)
        y = self.Op._rmatvec(x)
        if self.adj:
            if self.real:
                y = ncp.real(y)
            else:
                y = -ncp.imag(y)
        return y


def aslinearoperator(Op):
    """Return Op as a LinearOperator.

    Converts any operator into a LinearOperator. This can be used when `Op`
    is a private operator to ensure that the return operator has all properties
    and methods of the parent class.

    Parameters
    ----------
    Op : :obj:`pylops.LinearOperator` or any other Operator
        Operator of any type

    Returns
    -------
    Op : :obj:`pylops.LinearOperator`
        Operator of type :obj:`pylops.LinearOperator`

    """
    if isinstance(Op, LinearOperator):
        return Op
    else:
        return LinearOperator(Op)
